@@ -1,7 +1,7 @@
 %%
 %% %CopyrightBegin%
 %%
-%% Copyright Ericsson AB 1999-2020. All Rights Reserved.
+%% Copyright Ericsson AB 1999-2023. All Rights Reserved.
 %%
 %% Licensed under the Apache License, Version 2.0 (the "License");
 %% you may not use this file except in compliance with the License.
@@ -21,6 +21,7 @@
 -module(fun_SUITE).
 
 -export([all/0, suite/0,
+         init_per_testcase/2, end_per_testcase/2,
 	 bad_apply/1,bad_fun_call/1,badarity/1,ext_badarity/1,
          bad_arglist/1,
 	 equality/1,ordering/1,
@@ -46,6 +47,12 @@ all() ->
      const_propagation, t_arity, t_is_function2, t_fun_info,
      t_fun_info_mfa,t_fun_to_list].
 
+init_per_testcase(_TestCase, Config) ->
+    Config.
+end_per_testcase(_TestCase, Config) ->
+    erts_test_utils:ept_check_leaked_nodes(Config).
+
+
 %% Test that the correct EXIT code is returned for all types of bad funs.
 bad_apply(Config) when is_list(Config) ->
     bad_apply_fc(42, [0]),
@@ -574,7 +581,7 @@ refc_ets_bag(F1, Options) ->
     ok.
 
 refc_dist(Config) when is_list(Config) ->
-    {ok,Node} = start_node(fun_SUITE_refc_dist),
+    {ok, Peer, Node} = ?CT_PEER(),
     process_flag(trap_exit, true),
     Pid = spawn_link(Node, fun() -> receive
                                         Fun when is_function(Fun) ->
@@ -595,7 +602,7 @@ refc_dist(Config) when is_list(Config) ->
     true = erlang:garbage_collect(),
     2 = fun_refc(F),
     refc_dist_send(Node, F),
-    test_server:stop_node(Node).
+    peer:stop(Peer).
 
 refc_dist_send(Node, F) ->
     Pid = spawn_link(Node, fun() -> receive
@@ -652,7 +659,7 @@ fun_refc(F) ->
     Count.
 
 const_propagation(Config) when is_list(Config) ->
-    Fun1 = fun start_node/1,
+    Fun1 = fun wait_until/1,
     2 = fun_refc(Fun1),
     Fun2 = Fun1,
     my_cmp({Fun1,Fun2}),
@@ -678,14 +685,14 @@ t_arity(Config) when is_list(Config) ->
 
     %% Test that the arity is transferred properly.
     process_flag(trap_exit, true),
-    {ok,Node} = start_node(fun_test_arity),
+    {ok, Peer, Node} = ?CT_PEER(),
     hello_world = spawn_call(Node, fun() -> hello_world end),
     0 = spawn_call(Node, fun(X) -> X end),
     42 = spawn_call(Node, fun(_X) -> A end),
     43 = spawn_call(Node, fun(X, Y) -> A+X+Y end),
     1 = spawn_call(Node, fun(X, Y) -> X+Y end),
     45 = spawn_call(Node, fun(X, Y, Z) -> A+X+Y+Z end),
-    test_server:stop_node(Node),
+    peer:stop(Peer),
     ok.
 
 t_is_function2(Config) when is_list(Config) ->
@@ -868,12 +875,6 @@ fun_arity(F) ->
     {arity,Arity} = erlang:fun_info(F, arity),
     Arity.
 
-start_node(Name) ->
-    Pa = filename:dirname(code:which(?MODULE)),
-    Cookie = atom_to_list(erlang:get_cookie()),
-    test_server:start_node(Name, slave, 
-			   [{args, "-setcookie " ++ Cookie ++" -pa " ++ Pa}]).
-
 wait_until(Fun) ->
     case catch Fun() of
 	true -> ok;
