@@ -31,7 +31,8 @@
 	 start_timer/1, setup_timer/2, 
 	 reset_timer/1, cancel_timer/1,
          is_node_name/1, split_node/1, is_allowed/2,
-	 shutdown/3, shutdown/4]).
+	 shutdown/3, shutdown/4,
+         net_ticker_spawn_options/0]).
 
 -import(error_logger,[error_msg/2]).
 
@@ -44,15 +45,6 @@
 -define(shutdown_trace(A,B), noop).
 -endif.
 
--define(to_port(FSend, Socket, Data),
-	case FSend(Socket, Data) of
-	    {error, closed} ->
-		self() ! {tcp_closed, Socket},
-	        {error, closed};
-	    R ->
-	        R
-        end).
-
 
 -define(int16(X), [((X) bsr 8) band 16#ff, (X) band 16#ff]).
 
@@ -70,8 +62,6 @@
 -define(u32(X3,X2,X1,X0),
         (((X3) bsl 24) bor ((X2) bsl 16) bor ((X1) bsl 8) bor (X0))).
 
--define(CREATION_UNKNOWN,0).
-
 -record(tick, {read = 0,
 	       write = 0,
 	       tick = 0,
@@ -124,6 +114,10 @@ dflag2str(?DFLAG_FRAGMENTS) ->
     "FRAGMENTS";
 dflag2str(?DFLAG_HANDSHAKE_23) ->
     "HANDSHAKE_23";
+dflag2str(?DFLAG_UNLINK_ID) ->
+    "UNLINK_ID";
+dflag2str(?DFLAG_MANDATORY_25_DIGEST) ->
+    "MANDATORY_25_DIGEST";
 dflag2str(?DFLAG_SPAWN) ->
     "SPAWN";
 dflag2str(?DFLAG_NAME_ME) ->
@@ -132,8 +126,8 @@ dflag2str(?DFLAG_V4_NC) ->
     "V4_NC";
 dflag2str(?DFLAG_ALIAS) ->
     "ALIAS";
-dflag2str(_) ->
-    "UNKNOWN".
+dflag2str(Other) ->
+    lists:flatten(io_lib:format("UNKNOWN<~.16.0B>", [Other])).
 
 
 adjust_flags(ThisFlags, OtherFlags) ->
@@ -143,7 +137,7 @@ publish_flag(_, NameMeFlg, _) when (NameMeFlg band ?DFLAG_NAME_ME) =/= 0 ->
     ?DFLAG_NAME_ME;
 publish_flag(hidden, _, _) ->
     0;
-publish_flag(_, _, OtherNode) when is_atom(OtherNode) ->
+publish_flag(_, _, OtherNode) ->
     case net_kernel:publish_on_node(OtherNode) of
 	true ->
 	    ?DFLAG_PUBLISHED;
@@ -200,13 +194,13 @@ handshake_other_started(#hs_data{request_type=ReqType,
     AddFlgs = convert_flags(AddFlgs0),
     RejFlgs = convert_flags(RejFlgs0),
     ReqFlgs = convert_flags(ReqFlgs0),
-    {PreOtherFlags,NodeOrHost,Creation,SendNameVersion} = recv_name(HSData0),
+    {PreOtherFlags0,NodeOrHost,Creation} = recv_name(HSData0),
+    PreOtherFlags = expand_mandatory_25_flag(PreOtherFlags0),
     EDF = erts_internal:get_dflags(),
     PreThisFlags = make_this_flags(ReqType, AddFlgs, RejFlgs, NodeOrHost, EDF,
                                    PreOtherFlags),
     HSData1 = HSData0#hs_data{this_flags=PreThisFlags,
                               other_flags=PreOtherFlags,
-                              other_version=flags_to_version(PreOtherFlags),
                               other_node=NodeOrHost,
                               other_started=true,
                               other_creation=Creation,
@@ -215,19 +209,23 @@ handshake_other_started(#hs_data{request_type=ReqType,
                               require_flags=ReqFlgs},
     check_dflags(HSData1, EDF),
     ?debug({"MD5 connection from ~p~n", [NodeOrHost]}),
-    HSData2 = mark_pending(HSData1),
+    {AcceptedPending, HSData2} = mark_pending(HSData1),
     Node = HSData2#hs_data.other_node,
     Cookie = auth:get_cookie(Node),
     ChallengeA = gen_challenge(),
     send_challenge(HSData2, ChallengeA),
     reset_timer(HSData2#hs_data.timer),
-    HSData3 = recv_complement(HSData2, SendNameVersion),
+    HSData3 = recv_complement(HSData2),
     check_dflags(HSData3, EDF),
     ChosenFlags = adjust_flags(HSData3#hs_data.this_flags,
                                HSData3#hs_data.other_flags),
     HSData4 = HSData3#hs_data{this_flags = ChosenFlags,
                               other_flags = ChosenFlags},
     ChallengeB = recv_challenge_reply(HSData4, ChallengeA, Cookie),
+    case AcceptedPending of
+        up_pending -> wait_pending(HSData4);
+        _ -> continue
+    end,
     send_challenge_ack(HSData4, gen_digest(ChallengeB, Cookie)),
     ?debug({dist_util, self(), accept_connection, Node}),
     connection(HSData4);
@@ -235,6 +233,15 @@ handshake_other_started(#hs_data{request_type=ReqType,
 handshake_other_started(OldHsData) when element(1,OldHsData) =:= hs_data ->
     handshake_other_started(convert_old_hsdata(OldHsData)).
 
+expand_mandatory_25_flag(Flags) ->
+    if
+        Flags band ?DFLAG_MANDATORY_25_DIGEST =/= 0 ->
+            %% From OTP 25, the single flag ?DFLAG_MANDATORY_25_DIGEST can
+            %% replace all the flags in ?MANDATORY_DFLAGS_25.
+            Flags bor ?MANDATORY_DFLAGS_25;
+        true ->
+            Flags
+    end.
 
 %%
 %% Check mandatory flags...
@@ -242,10 +249,19 @@ handshake_other_started(OldHsData) when element(1,OldHsData) =:= hs_data ->
 check_dflags(#hs_data{other_node = Node,
                       other_flags = OtherFlags,
                       other_started = OtherStarted,
-                      require_flags = RequiredFlags} = HSData,
+                      require_flags = RequiredFlags,
+                      other_creation = OtherCreation} = HSData,
              #erts_dflags{}=EDF) ->
 
-    Mandatory = (EDF#erts_dflags.mandatory bor RequiredFlags),
+    Mask = case OtherCreation of
+               undefined ->
+                   %% Old 'send_name' without creation and high flag bits
+                   %% Only check low 32 flag bits for now
+                   (1 bsl 32) - 1;
+               _ ->
+                   bnot 0
+           end,
+    Mandatory = Mask band (EDF#erts_dflags.mandatory bor RequiredFlags),
     Missing = check_mandatory(Mandatory, OtherFlags, []),
     case Missing of
         [] ->
@@ -288,8 +304,10 @@ check_mandatory(Mandatory, OtherFlags, Missing) ->
 mark_pending(#hs_data{kernel_pid=Kernel,
 		      other_node=Node,
 		      this_node=MyNode}=HSData) ->
-    case do_mark_pending(Kernel, MyNode, Node,
-			 HSData#hs_data.other_flags) of
+    KernelReply = do_mark_pending(Kernel, MyNode, Node,
+                                  HSData#hs_data.other_flags),
+    {KernelReply,
+     case KernelReply of
 	ok ->
 	    send_status(HSData, ok),
 	    reset_timer(HSData#hs_data.timer),
@@ -319,8 +337,12 @@ mark_pending(#hs_data{kernel_pid=Kernel,
 
 	    %% This can happen if the other node goes down,
 	    %% and goes up again and contact us before we have
-	    %% detected that the socket was closed. 
-	    wait_pending(Kernel),
+	    %% detected that the socket was closed.
+            %% It can also happen if the old connection went down silently,
+            %% without us knowing, a lost TCP FIN or RST packet for example.
+
+            %% Continue handshake to verify cookie and then wait for old
+            %% connection to die.
 	    reset_timer(HSData#hs_data.timer),
             HSData;
 
@@ -328,15 +350,17 @@ mark_pending(#hs_data{kernel_pid=Kernel,
 	    %% FIXME: is this a case ?
 	    ?debug({dist_util,self(),mark_pending,already_pending,Node}),
 	    ?shutdown(Node)
-    end.
+        end
+    }.
+
 
 
 %%
-%% Marking pending and negotiating away 
-%% simultaneous connection problems
+%% Tell net_kernel we are waiting for old connection to die.
 %%
-
-wait_pending(Kernel) ->
+wait_pending(#hs_data{kernel_pid=Kernel,
+		      other_node=Node}) ->
+    Kernel ! {self(), {wait_pending, Node}},
     receive
 	{Kernel, pending} ->
 	    ?trace("wait_pending returned for pid ~p.~n", 
@@ -388,7 +412,7 @@ shutdown(_Module, _Line, _Data, Reason) ->
     exit(Reason).
 %% Use this line to debug connection.  
 %% Set net_kernel verbose = 1 as well.
-%%    exit({Reason, ?MODULE, _Line, _Data, erlang:timestamp()}).
+%%    exit({Reason, {?MODULE, _Line, _Data, erlang:timestamp()}}).
 
 handshake_we_started(#hs_data{request_type=ReqType,
                               this_node=MyNode,
@@ -400,36 +424,33 @@ handshake_we_started(#hs_data{request_type=ReqType,
     RejFlgs = convert_flags(RejFlgs0),
     ReqFlgs = convert_flags(ReqFlgs0),
     EDF = erts_internal:get_dflags(),
-    {NameMeFlg, NameToSend, MinVersion} =
+    {NameMeFlg, NameToSend} =
         case node() of
             nonode@nohost ->
                 {node, "undefined", Host} = split_node(MyNode),
-                false = net_kernel:dist_listen(),
-                {?DFLAG_NAME_ME, Host, ?ERL_DIST_VER_6};
+                {?DFLAG_NAME_ME, Host};
 
             _ ->
-                {0, MyNode, PreHSData#hs_data.other_version}
+                {0, MyNode}
         end,
     PreThisFlags = make_this_flags(ReqType, AddFlgs, RejFlgs, Node, EDF, NameMeFlg),
     HSData = PreHSData#hs_data{this_node = NameToSend,
                                this_flags = PreThisFlags,
-                               other_version = MinVersion,
                                add_flags = AddFlgs,
                                reject_flags = RejFlgs,
                                require_flags = ReqFlgs},
-    SendNameVersion = send_name(HSData),
+    send_name(HSData),
     HSData1 = recv_status(HSData),
-    {PreOtherFlags, ChallengeA, Creation} = recv_challenge(HSData1),
+    {PreOtherFlags0, ChallengeA, Creation} = recv_challenge(HSData1),
+    PreOtherFlags = expand_mandatory_25_flag(PreOtherFlags0),
     ChosenFlags = adjust_flags(PreThisFlags, PreOtherFlags),
     HSData2 = HSData1#hs_data{this_flags = ChosenFlags,
 			       other_flags = ChosenFlags,
 			       other_started = false,
-                               other_version = flags_to_version(PreOtherFlags),
                                other_creation = Creation},
     check_dflags(HSData2, EDF),
     ChallengeB = gen_challenge(),
     Cookie = auth:get_cookie(Node),
-    send_complement(HSData2, SendNameVersion),
     send_challenge_reply(HSData2, ChallengeB,
                          gen_digest(ChallengeA, Cookie)),
     reset_timer(HSData2#hs_data.timer),
@@ -450,16 +471,6 @@ convert_flags(Flags) when is_integer(Flags) ->
 convert_flags(_Undefined) ->
     0.
 
-flags_to_version(Flags) ->
-    case Flags band ?DFLAG_HANDSHAKE_23 of
-        0 ->
-            ?ERL_DIST_VER_5;
-        ?DFLAG_HANDSHAKE_23 ->
-            ?ERL_DIST_VER_6
-    end.
-
-
-
 %% --------------------------------------------------------------
 %% The connection has been established.
 %% --------------------------------------------------------------
@@ -504,11 +515,13 @@ connection(#hs_data{other_node = Node,
                                     f_setopts = HSData#hs_data.mf_setopts,
                                     f_getopts = HSData#hs_data.mf_getopts},
 			     #tick{});
-		_ ->
-		    ?shutdown2(Node, connection_setup_failed)
+		Error1 ->
+		    ?shutdown2(
+                       {Node, Socket},
+                       {f_setopts_post_nodeup_failed, Error1})
 	    end;
-	_ ->
-	    ?shutdown(Node)
+	Error2 ->
+	    ?shutdown2({Node, Socket}, {f_setopts_pre_nodeup_failed, Error2})
     end.
 
 %% Generate a message digest from Challenge number and Cookie	
@@ -536,7 +549,7 @@ gen_challenge() ->
 %% No error return; either succeeds or terminates the process.
 do_setnode(#hs_data{other_node = Node, socket = Socket, 
                     this_node = MyNode,
-		    other_flags = Flags, other_version = Version,
+		    other_flags = Flags,
 		    f_getll = GetLL,
                     other_creation = Creation}=HSData) ->
     case GetLL(Socket) of
@@ -549,19 +562,19 @@ do_setnode(#hs_data{other_node = Node, socket = Socket,
                           MyNode ->
                               false
                       end,
-	    ?trace("setnode: node=~p port=~p flags=~p(~p) ver=~p creation=~p~n",
-		   [Node, Port, Flags, publish_type(Flags), Version, Creation]),
+	    ?trace("setnode: node=~p port=~p flags=~p(~p) creation=~p~n",
+		   [Node, Port, Flags, publish_type(Flags), Creation]),
 
             MyNode = node(), % ASSERT
             try
-                DHandle = erlang:setnode(Node, Port, {Flags, Version, Creation}),
+                DHandle = erlang:setnode(Node, Port, {Flags, Creation}),
                 {DHandle, NamedMe}
             catch
                 error:system_limit ->
 		    error_msg("** Distribution system limit reached, "
 			      "no table space left for node ~w ** ~n",
 			      [Node]),
-		    ?shutdown(Node);
+		    ?shutdown({Node, Socket});
                 error:Other:Stacktrace ->
                     exit({Other, Stacktrace})
 	    end;
@@ -569,7 +582,7 @@ do_setnode(#hs_data{other_node = Node, socket = Socket,
 	    error_msg("** Distribution connection error, "
 		      "could not get low level port for node ~w ** ~n",
 		      [Node]),
-	    ?shutdown(Node)
+	    ?shutdown({Node, Socket})
     end.
 
 mark_nodeup(#hs_data{kernel_pid = Kernel, 
@@ -606,9 +619,9 @@ con_loop(#state{kernel = Kernel, node = Node,
 	 Tick) ->
     receive
 	{tcp_closed, Socket} ->
-	    ?shutdown2(Node, connection_closed);
+	    ?shutdown2({Node, Socket}, tcp_closed);
 	{Kernel, disconnect} ->
-	    ?shutdown2(Node, disconnected);
+	    ?shutdown2({Node, Socket}, disconnected);
 	{Kernel, aux_tick} ->
 	    case getstat(DHandle, Socket, MFGetstat) of
 		{ok, _, _, PendWrite} ->
@@ -625,17 +638,17 @@ con_loop(#state{kernel = Kernel, node = Node,
  		    error_msg("** Node ~p not responding **~n"
  			      "** Removing (timedout) connection **~n",
  			      [Node]),
- 		    ?shutdown2(Node, net_tick_timeout);
-		_Other ->
-		    ?shutdown2(Node, send_net_tick_failed)
+		    ?shutdown2({Node, Socket}, net_tick_timeout);
+		Error1 ->
+		    ?shutdown2({Node, Socket}, {send_net_tick_failed, Error1})
 	    end;
 	{From, get_status} ->
 	    case getstat(DHandle, Socket, MFGetstat) of
 		{ok, Read, Write, _} ->
 		    From ! {self(), get_status, {ok, Read, Write}},
 		    con_loop(ConData, Tick);
-		_ ->
-		    ?shutdown2(Node, get_status_failed)
+		Error2 ->
+		    ?shutdown2({Node, Socket}, {get_status_failed, Error2})
 	    end;
 	{From, Ref, {setopts, Opts}} ->
 	    Ret = case MFSetOpts of
@@ -660,35 +673,17 @@ con_loop(#state{kernel = Kernel, node = Node,
 
 send_name(#hs_data{socket = Socket, this_node = Node, 
 		   f_send = FSend, 
-		   this_flags = Flags,
-		   other_version = Version}) ->
+		   this_flags = Flags}) ->
     NameBin = to_binary(Node),
-    if Version =:= undefined;
-       Version =:= ?ERL_DIST_VER_5 ->
-            %% We treat "5" the same as 'undefined' as there are
-            %% custom made epmd modules out there with a hardcoded "5".
-            %%
-            %% Send old 'n' message but with DFLAG_HANDSHAKE_23
-            %% Old nodes will ignore DFLAG_HANDSHAKE_23 and reply old 'n' challenge.
-            %% New nodes will see DFLAG_HANDSHAKE_23 and reply new 'N' challenge.
-            ?trace("send_name: 'n' node=~p, version=~w\n",
-                   [Node, ?ERL_DIST_VER_5]),
-            _ = ?to_port(FSend, Socket,
-                         [<<$n, ?ERL_DIST_VER_5:16, Flags:32>>, NameBin]),
-            ?ERL_DIST_VER_5;
-
-       is_integer(Version), Version >= ?ERL_DIST_VER_6 ->
-            Creation = case name_type(Flags) of
-                           static -> erts_internal:get_creation();
-                           dynamic -> 0
-                       end,
-            NameLen = byte_size(NameBin),
-            ?trace("send_name: 'N' node=~p creation=~w\n",
-                   [Node, Creation]),
-            _ = ?to_port(FSend, Socket,
-                         [<<$N, Flags:64, Creation:32, NameLen:16>>, NameBin]),
-            ?ERL_DIST_VER_6
-    end.
+    Creation = case name_type(Flags) of
+                   static -> erts_internal:get_creation();
+                   dynamic -> 0
+               end,
+    NameLen = byte_size(NameBin),
+    ?trace("send_name: 'N' node=~p creation=~w\n",
+           [Node, Creation]),
+    to_port(FSend, Socket,
+            [<<$N, Flags:64, Creation:32, NameLen:16>>, NameBin]).
 
 to_binary(Atom) when is_atom(Atom) ->
     atom_to_binary(Atom, latin1);
@@ -697,61 +692,28 @@ to_binary(List) when is_list(List) ->
 
 send_challenge(#hs_data{socket = Socket, this_node = Node, 
 			this_flags = ThisFlags,
-                        other_flags = OtherFlags,
 			f_send = FSend},
 	       Challenge ) ->
-    case OtherFlags band ?DFLAG_HANDSHAKE_23 of
-        0 ->
-            %% Reply with old 'n' message
-            ?trace("send: 'n' challenge=~w\n", [Challenge]),
-
-            ?to_port(FSend, Socket, [<<$n,
-                                       ?ERL_DIST_VER_5:16, % echo same Version back
-                                       ThisFlags:32,
-                                       Challenge:32>>,
-                                     atom_to_list(Node)]);
-
-        ?DFLAG_HANDSHAKE_23 ->
-            %% Reply with new 'N' message
-            Creation = erts_internal:get_creation(),
-            NodeName = atom_to_binary(Node, latin1),
-            NameLen = byte_size(NodeName),
-            ?trace("send: 'N' challenge=~w creation=~w\n",
-                   [Challenge,Creation]),
-            ?to_port(FSend, Socket, [<<$N,
-                                       ThisFlags:64,
-                                       Challenge:32,
-                                       Creation:32,
-                                       NameLen:16>>, NodeName])
-    end.
-
-send_complement(#hs_data{socket = Socket,
-                         f_send = FSend,
-                         this_flags = Flags,
-                         other_flags = Flags},
-                SendNameVersion) ->
-    if SendNameVersion =:= ?ERL_DIST_VER_5,
-       (Flags band ?DFLAG_HANDSHAKE_23) =/= 0 ->
-            %% We sent an old 'n' name message and need to complement
-            %% with creation value.
-            Creation = erts_internal:get_creation(),
-            FlagsHigh = Flags bsr 32,
-            ?trace("send_complement: 'c' flags_high=~w creation=~w\n", [FlagsHigh,Creation]),
-            ?to_port(FSend, Socket, [<<$c, FlagsHigh:32, Creation:32>>]);
-       true->
-            ok % no complement msg needed
-    end.
+    %% Reply with new 'N' message
+    Creation = erts_internal:get_creation(),
+    NodeName = atom_to_binary(Node, latin1),
+    NameLen = byte_size(NodeName),
+    ?trace("send: 'N' challenge=~w creation=~w\n",
+           [Challenge,Creation]),
+    to_port(FSend, Socket,
+            [<<$N,ThisFlags:64, Challenge:32, Creation:32, NameLen:16>>,
+             NodeName]).
 
 send_challenge_reply(#hs_data{socket = Socket, f_send = FSend}, 
 		     Challenge, Digest) ->
     ?trace("send_reply: challenge=~w digest=~p\n",
 	   [Challenge,Digest]),
-    ?to_port(FSend, Socket, [$r,?int32(Challenge),Digest]).
+    to_port(FSend, Socket, [$r,?int32(Challenge),Digest]).
 
 send_challenge_ack(#hs_data{socket = Socket, f_send = FSend}, 
 		   Digest) ->
     ?trace("send_ack: digest=~p\n", [Digest]),
-    ?to_port(FSend, Socket, [$a,Digest]).
+    to_port(FSend, Socket, [$a,Digest]).
 
 
 %%
@@ -764,10 +726,19 @@ recv_name(#hs_data{socket = Socket, f_recv = Recv} = HSData) ->
             recv_name_old(HSData, Data);
         {ok, [$N | _] = Data} ->
             recv_name_new(HSData, Data);
-	_ ->
-	    ?shutdown(no_node)
+	Other ->
+	    ?shutdown2({no_node, Socket}, {recv_name_failed, Other})
     end.
 
+%% OTP 25.3:
+%% We accept old 'send_name' messages ($n) used by OTP-22 and older.
+%% OTP-23 or OTP-24 nodes, not using epmd, may send the old message
+%% to us if they do not know our OTP version.
+%% Hence, we accept the old 'send_name' and the accompanying
+%% 'send_complement', but we do not send them ourself.
+%% This was removed prematurely in OTP-25.0,
+%% therefore versions 25.0 to 25.2.* are broken in this regard.
+%% Can be safely removed in OTP-27.0.
 recv_name_old(HSData,
              [$n, V1, V0, F3, F2, F1, F0 | Node] = Data) ->
     <<_Version:16>> = <<V1,V0>>,
@@ -776,7 +747,7 @@ recv_name_old(HSData,
     case is_node_name(Node) of
         true ->
             check_allowed(HSData, Node),
-            {Flags, list_to_atom(Node), ?CREATION_UNKNOWN, ?ERL_DIST_VER_5};
+            {Flags, list_to_atom(Node), undefined};
         false ->
             ?shutdown(Data)
     end.
@@ -799,9 +770,9 @@ recv_name_new(HSData,
                                  %% Keep host name as string
                                  Name
                          end,
-            {Flags, NodeOrHost, Creation, ?ERL_DIST_VER_6};
+            {Flags, NodeOrHost, Creation};
         false ->
-            ?shutdown(Data)
+            ?shutdown({name, Data})
     end.
 
 is_node_name(NodeName) ->
@@ -934,33 +905,12 @@ publish_type(Flags) ->
 %% wait for challenge after connect
 recv_challenge(#hs_data{socket=Socket, f_recv=Recv}=HSData) ->
     case Recv(Socket, 0, infinity) of
-	{ok, [$n | _]=Msg} ->
-            recv_challenge_old(HSData, Msg);
         {ok,[$N | _]=Msg} ->
             recv_challenge_new(HSData, Msg);
 	Other ->
-            ?shutdown2(no_node, {recv_challenge_failed, Other})
+            ?shutdown2({no_node, Socket}, {recv_challenge_failed, Other})
     end.
 
-recv_challenge_old(#hs_data{other_node=Node},
-                   [$n, V1,V0, F3,F2,F1,F0, C3,C2,C1,C0 | Ns]=Msg) ->
-    <<_Version:16>> = <<V1,V0>>,
-    <<Flags:32>> = <<F3,F2,F1,F0>>,
-    <<Challenge:32>> = <<C3,C2,C1,C0>>,
-    ?trace("recv: 'n' node=~p, challenge=~w version=~w\n",
-           [Ns, Challenge, _Version]),
-    try {list_to_existing_atom(Ns), Flags band ?DFLAG_HANDSHAKE_23} of
-        {Node, 0} ->
-            {Flags, Challenge, ?CREATION_UNKNOWN};
-        _ ->
-            ?shutdown2(no_node, {recv_challenge_failed, version, Msg})
-    catch
-        error:badarg ->
-            ?shutdown2(no_node, {recv_challenge_failed, no_node, Ns})
-    end;
-recv_challenge_old(_, Other) ->
-    ?shutdown2(no_node, {recv_challenge_failed, Other}).
-
 recv_challenge_new(#hs_data{other_node=Node},
                    [$N,
                     F7,F6,F5,F4,F3,F2,F1,F0,
@@ -981,43 +931,36 @@ recv_challenge_new(#hs_data{other_node=Node},
     ?trace("recv: 'N' node=~p, challenge=~w creation=~w\n",
            [Ns, Challenge, Creation]),
 
-    case Flags band ?DFLAG_HANDSHAKE_23 of
-        ?DFLAG_HANDSHAKE_23 ->
-            try list_to_existing_atom(Ns) of
-                Node ->
-                    {Flags, Challenge, Creation};
-                _ ->
-                    ?shutdown2(no_node, {recv_challenge_failed, no_node, Ns})
-            catch
-                error:badarg ->
-                    ?shutdown2(no_node, {recv_challenge_failed, no_node, Ns})
-            end;
-       0 ->
-            ?shutdown2(no_node, {recv_challenge_failed, version, Msg})
+    try list_to_existing_atom(Ns) of
+        Node ->
+            {Flags, Challenge, Creation};
+        _ ->
+            ?shutdown2(no_node, {recv_challenge_failed, no_node, Ns})
+    catch
+        error:badarg ->
+            ?shutdown2(no_node, {recv_challenge_failed, no_node, Ns})
     end;
 recv_challenge_new(_, Other) ->
     ?shutdown2(no_node, {recv_challenge_failed, Other}).
 
 
+%% See comment for recv_name_old
 recv_complement(#hs_data{socket = Socket,
                          f_recv = Recv,
-                         other_flags = Flags} = HSData,
-                SendNameVersion) ->
-    if SendNameVersion =:= ?ERL_DIST_VER_5,
-       (Flags band ?DFLAG_HANDSHAKE_23) =/= 0 ->
-            case Recv(Socket, 0, infinity) of
-                {ok, [$c, F7,F6,F5,F4, Cr3,Cr2,Cr1,Cr0]} ->
-                    <<FlagsHigh:32>> = <<F7,F6,F5,F4>>,
-                    <<Creation:32>> = <<Cr3,Cr2,Cr1,Cr0>>,
-                    ?trace("recv_complement: creation=~w\n", [Creation]),
-                    HSData#hs_data{other_creation = Creation,
-                                   other_flags = Flags bor (FlagsHigh bsl 32)};
-                Other ->
-                    ?shutdown2(no_node, {recv_complement_failed, Other})
-            end;
-         true ->
-            HSData
-    end.
+                         other_flags = Flags,
+                         other_creation = undefined} = HSData) ->
+    case Recv(Socket, 0, infinity) of
+        {ok, [$c, F7,F6,F5,F4, Cr3,Cr2,Cr1,Cr0]} ->
+            <<FlagsHigh:32>> = <<F7,F6,F5,F4>>,
+            <<Creation:32>> = <<Cr3,Cr2,Cr1,Cr0>>,
+            ?trace("recv_complement: creation=~w\n", [Creation]),
+            HSData#hs_data{other_creation = Creation,
+                           other_flags = Flags bor (FlagsHigh bsl 32)};
+        Other ->
+            ?shutdown2(no_node, {recv_complement_failed, Other})
+    end;
+recv_complement(HSData) ->
+    HSData.
 
 
 %%
@@ -1043,7 +986,8 @@ recv_challenge_reply(#hs_data{socket = Socket,
 		    ?shutdown2(NodeB, {recv_challenge_reply_failed, bad_cookie})
 	    end;
 	Other ->
-	    ?shutdown2(no_node, {recv_challenge_reply_failed, Other})
+	    ?shutdown2({no_node, Socket},
+                       {recv_challenge_reply_failed, Other})
     end.
 
 recv_challenge_ack(#hs_data{socket = Socket, f_recv = FRecv, 
@@ -1060,15 +1004,21 @@ recv_challenge_ack(#hs_data{socket = Socket, f_recv = FRecv,
 		_ ->
 		    error_msg("** Connection attempt to node ~w cancelled."
                               " Invalid challenge ack. **~n", [NodeB]),
-		    ?shutdown2(NodeB, {recv_challenge_ack_failed, bad_cookie})
+		    ?shutdown2(
+                       {NodeB, Socket}, {recv_challenge_ack_failed, bad_cookie})
 	    end;
 	Other ->
-	    ?shutdown2(NodeB, {recv_challenge_ack_failed, Other})
+	    ?shutdown2(
+               {NodeB, Socket}, {recv_challenge_ack_failed, Other})
     end.
 
-recv_status(#hs_data{kernel_pid = Kernel, socket = Socket, 
-                     this_flags = MyFlgs,
-		     other_node = Node, f_recv = Recv} = HSData) ->
+recv_status(
+  #hs_data{
+     kernel_pid = Kernel,
+     socket = Socket,
+     this_flags = MyFlgs,
+     other_node = Node,
+     f_recv = Recv} = HSData) ->
     case Recv(Socket, 0, infinity) of
         {ok, "snamed:"++Rest} ->
             <<NameLen:16,
@@ -1085,9 +1035,11 @@ recv_status(#hs_data{kernel_pid = Kernel, socket = Socket,
 	    ?debug({dist_util,self(),recv_status, Node, Stat}),
 	    case {Stat, name_type(MyFlgs)} of
                 {"not_allowed", _} ->
-                    ?shutdown2(Node, {recv_status_failed, not_allowed});
+                    ?shutdown2(
+                       {Node, Socket}, {recv_status_failed, not_allowed});
                 {_, dynamic} ->
-                    ?shutdown2(Node, {recv_status_failed, unexpected, Stat});
+                    ?shutdown2(
+                       {Node, Socket}, {recv_status_failed, unexpected, Stat});
                 _ ->
                     continue
             end,
@@ -1102,20 +1054,20 @@ recv_status(#hs_data{kernel_pid = Kernel, socket = Socket,
                     ?debug({is_pending,self(),Reply}),
                     send_status(HSData, Reply),
                     if not Reply ->
-                            ?shutdown(Node);
+                            ?shutdown({Node, Socket});
                        Reply ->
                             HSData
                     end;
                 "ok" -> HSData;
                 "ok_simultaneous" -> HSData;
                 Other ->
-                    ?shutdown2(Node, {recv_status_failed, unknown, Other})
+                    ?shutdown2(
+                       {Node, Socket}, {recv_status_failed, unknown, Other})
             end;
 
 	Error ->
-	    ?debug({dist_util,self(),recv_status_error, 
-		Node, Error}),
-	    ?shutdown2(Node, {recv_status_failed, Error})
+	    ?debug({dist_util, self(), recv_status_error, Node, Error}),
+	    ?shutdown2({Node, Socket}, {recv_status_failed, Error})
     end.
 
 
@@ -1129,13 +1081,14 @@ recv_status_reply(#hs_data{socket = Socket,
                 "true" -> true;
                 "false" -> false;
                 Other ->
-                    ?shutdown2(Node, {recv_status_failed, unexpected, Other})
+                    ?shutdown2(
+                       {Node, Socket}, {recv_status_failed, unexpected, Other})
             end;
 
 	Error ->
 	    ?debug({dist_util,self(),recv_status_error,
 		Node, Error}),
-	    ?shutdown2(Node, {recv_status_failed, Error})
+	    ?shutdown2({Node, Socket}, {recv_status_failed, Error})
     end.
 
 
@@ -1150,22 +1103,32 @@ send_status(#hs_data{socket = Socket,
     case FSend(Socket, [$s, "named:",
                        <<NameLen:16, NameBin/binary>>,
                        <<Creation:32>>]) of
-	{error, _} ->
-	    ?shutdown(Node);
+	{error, _} = Error->
+	    ?shutdown2({Node, Socket}, {send_status_failed, Error});
 	_ ->
-	    true
+	    ok
     end;
 send_status(#hs_data{socket = Socket, other_node = Node,
 		    f_send = FSend}, Stat) ->
     ?debug({dist_util,self(),send_status, Node, Stat}),
     case FSend(Socket, [$s | atom_to_list(Stat)]) of
-	{error, _} ->
-	    ?shutdown(Node);
-	_ -> 
-	    true
+	{error, _} = Error ->
+	    ?shutdown2({Node, Socket}, {send_status_failed, Error});
+	_ ->
+	    ok
     end.
-    
-    
+
+to_port(FSend, Socket, Data) ->
+    case FSend(Socket, Data) of
+        {error, closed} ->
+            self() ! {tcp_closed, Socket},
+            ok;
+        {error, _} = Error ->
+            ?shutdown2(Socket, {f_send_failed, Error});
+        ok ->
+            ok
+    end.
+
 
 %%
 %% Send a TICK to the other side.
@@ -1181,8 +1144,8 @@ send_status(#hs_data{socket = Socket, other_node = Node,
 %% ticker process, In that case this code will never run 
 
 %% And then every 60 seconds we also check the connection and 
-%% close it if we havn't received anything on it for the 
-%% last 60 secs. If ticked == tick we havn't received anything 
+%% close it if we haven't received anything on it for the 
+%% last 60 secs. If ticked == tick we haven't received anything 
 %% on the connection the last 60 secs. 
 
 %% The detection time interval is thus, by default, 45s < DT < 75s 
@@ -1260,7 +1223,7 @@ setup_timer(Pid, Timeout) ->
 	    setup_timer(Pid, Timeout)
     after Timeout ->
 	    ?trace("Timer expires ~p, ~p~n",[Pid, Timeout]),
-	    ?shutdown2(timer, setup_timer_timeout)
+	    ?shutdown2({timer, Pid}, setup_timer_timeout)
     end.
 
 reset_timer(Timer) ->
@@ -1269,5 +1232,22 @@ reset_timer(Timer) ->
 
 cancel_timer(Timer) ->
     unlink(Timer),
-    exit(Timer, shutdown).
+    exit(Timer, cancel_setup_timer).
+
+net_ticker_spawn_options() ->
+    Opts = application:get_env(kernel, net_ticker_spawn_options, []),
+    [link, {priority, max} | cleanup_net_ticker_spawn_options(Opts)].
+
+cleanup_net_ticker_spawn_options([]) ->
+    [];
+cleanup_net_ticker_spawn_options([link|Opts]) ->
+    cleanup_net_ticker_spawn_options(Opts);
+cleanup_net_ticker_spawn_options([{priority, _}|Opts]) ->
+    cleanup_net_ticker_spawn_options(Opts);
+cleanup_net_ticker_spawn_options([monitor|Opts]) ->
+    cleanup_net_ticker_spawn_options(Opts);
+cleanup_net_ticker_spawn_options([{monitor, _}|Opts]) ->
+    cleanup_net_ticker_spawn_options(Opts);
+cleanup_net_ticker_spawn_options([Opt|Opts]) ->
+    [Opt|cleanup_net_ticker_spawn_options(Opts)].
 
