@@ -1,7 +1,7 @@
 %%
 %% %CopyrightBegin%
 %%
-%% Copyright Ericsson AB 1996-2018. All Rights Reserved.
+%% Copyright Ericsson AB 1996-2023. All Rights Reserved.
 %%
 %% Licensed under the Apache License, Version 2.0 (the "License");
 %% you may not use this file except in compliance with the License.
@@ -25,10 +25,10 @@
 -export([all/0, suite/0,groups/0,init_per_group/2,end_per_group/2]).
 -export([set_path/1, get_path/1, add_path/1, add_paths/1, del_path/1,
 	 replace_path/1, load_file/1, load_abs/1, ensure_loaded/1,
-	 delete/1, purge/1, purge_many_exits/1, soft_purge/1, is_loaded/1,
-	 all_loaded/1,
+	 delete/1, purge/1, purge_many_exits/0, purge_many_exits/1,
+         soft_purge/1, is_loaded/1, all_loaded/1, all_available/1,
 	 load_binary/1, dir_req/1, object_code/1, set_path_file/1,
-	 upgrade/1,
+	 upgrade/0, upgrade/1,
 	 sticky_dir/1, pa_pz_option/1, add_del_path/1,
 	 dir_disappeared/1, ext_mod_dep/1, clash/1,
 	 where_is_file/1,
@@ -40,8 +40,8 @@
 	 on_load_deleted/1,
 	 big_boot_embedded/1,
          module_status/1,
-	 native_early_modules/1, get_mode/1,
-	 normalized_paths/1]).
+	 get_mode/1, code_path_cache/1,
+	 normalized_paths/1, mult_embedded_flags/1]).
 
 -export([init_per_testcase/2, end_per_testcase/2,
 	 init_per_suite/1, end_per_suite/1]).
@@ -55,14 +55,14 @@
 
 suite() ->
     [{ct_hooks,[ts_install_cth]},
-     {timetrap,{minutes,5}}].
+     {timetrap,{seconds,30}}].
 
 all() ->
     [set_path, get_path, add_path, add_paths, del_path,
      replace_path, load_file, load_abs, ensure_loaded,
      delete, purge, purge_many_exits, soft_purge, is_loaded, all_loaded,
-     load_binary, dir_req, object_code, set_path_file,
-     upgrade,
+     all_available, load_binary, dir_req, object_code, set_path_file,
+     upgrade, code_path_cache,
      sticky_dir, pa_pz_option, add_del_path, dir_disappeared,
      ext_mod_dep, clash, where_is_file,
      purge_stacktrace, mult_lib_roots,
@@ -72,7 +72,8 @@ all() ->
      on_load_purge, on_load_self_call, on_load_pending,
      on_load_deleted,
      module_status,
-     big_boot_embedded, native_early_modules, get_mode, normalized_paths].
+     big_boot_embedded, get_mode, normalized_paths,
+     mult_embedded_flags].
 
 %% These need to run in order
 groups() -> [{sequence, [sequence], [on_load_update,
@@ -140,6 +141,11 @@ end_per_testcase(on_load_embedded, Config) ->
     LinkName = proplists:get_value(link_name, Config),
     _ = del_link(LinkName),
     end_per_testcase(Config);
+end_per_testcase(upgrade, Config) ->
+    %% Make sure tracing is turned off even if the test times out.
+    erlang:trace_pattern({error_handler,undefined_function,3}, false, [global]),
+    erlang:trace(self(), false, [call]),
+    end_per_testcase(Config);
 end_per_testcase(_Func, Config) ->
     end_per_testcase(Config).
 
@@ -297,7 +303,7 @@ replace_path(Config) when is_list(Config) ->
     true = code:set_path(P),			%Reset path
     ok = file:del_dir("./kernel-2.11"),
 
-    %% Add a completly new application.
+    %% Add a completely new application.
 
     NewAppName = 'blurf_blarfer',
     NewAppDir = filename:join(Cwd, atom_to_list(NewAppName) ++ "-6.33.1"),
@@ -310,6 +316,68 @@ replace_path(Config) when is_list(Config) ->
 
     ok.
 
+code_path_cache(Config) ->
+    PrivDir = proplists:get_value(priv_dir, Config),
+
+    non_existing = code:which(?TESTMOD), % verify dummy name not in path
+    code:purge(?TESTMOD), % ensure no previous version in memory
+    code:delete(?TESTMOD),
+    code:purge(?TESTMOD),
+
+    Original = code:get_path(),
+    Dir = filename:join(PrivDir, "myebin"),
+    filelib:ensure_path(Dir),
+    ToBeReplacedDir1 = filename:join(PrivDir, "tobereplaced-1/ebin"),
+    ToBeReplacedDir2 = filename:join(PrivDir, "tobereplaced-2/ebin"),
+    filelib:ensure_path(ToBeReplacedDir1),
+    filelib:ensure_path(ToBeReplacedDir2),
+
+    File = filename:join(Dir, ?TESTMODOBJ),
+    {ok,?TESTMOD,Bin} = compile:forms(dummy_ast(), []),
+
+    %% Adding a cached path and re-adding for clearing
+    [begin
+        error = code:get_object_code(?TESTMOD),
+        true = code:Fun(Dir, cache),
+        error = code:get_object_code(?TESTMOD),
+        ok = file:write_file(File, Bin),
+        error = code:get_object_code(?TESTMOD),
+        true = code:Fun(Dir, cache),
+        {_,_,_} = code:get_object_code(?TESTMOD),
+        code:del_path(Dir),
+        ok = file:delete(File)
+     end || Fun <- [add_path, add_patha, add_pathz]],
+
+    Original = code:get_path(),
+
+    %% Adding several cached paths and explicit cache clearing
+    [begin
+        error = code:get_object_code(?TESTMOD),
+        ok = code:Fun([Dir, ToBeReplacedDir1], cache),
+        error = code:get_object_code(?TESTMOD),
+        ok = file:write_file(File, Bin),
+        error = code:get_object_code(?TESTMOD),
+        ok = code:clear_cache(),
+        {_,_,_} = code:get_object_code(?TESTMOD),
+        ok = code:del_paths([Dir, ToBeReplacedDir1]),
+        ok = file:delete(File)
+     end || Fun <- [add_paths, add_pathsa, add_pathsz]],
+
+    Original = code:get_path(),
+
+    %% Replacing a non-cached path with cache and set_path for clearing
+    error = code:get_object_code(?TESTMOD),
+    true = code:add_path(ToBeReplacedDir1),
+    true = code:replace_path(tobereplaced, ToBeReplacedDir2, cache),
+    error = code:get_object_code(?TESTMOD),
+    ok = file:write_file(filename:join(ToBeReplacedDir2, ?TESTMODOBJ), Bin),
+    error = code:get_object_code(?TESTMOD),
+    true = code:set_path(code:get_path(), nocache),
+    {_,_,_} = code:get_object_code(?TESTMOD),
+
+    true = code:set_path(Original),
+    ok.
+
 %% OTP-3977.
 dir_disappeared(Config) when is_list(Config) ->
     PrivDir = proplists:get_value(priv_dir, Config),
@@ -349,7 +417,7 @@ load_abs(Config) when is_list(Config) ->
 ensure_loaded(Config) when is_list(Config) ->
     {module, lists} = code:ensure_loaded(lists),
     case init:get_argument(mode) of
-	{ok, [["embedded"]]} ->
+	{ok, [["embedded"] | _]} ->
 	    {error, embedded} = code:ensure_loaded(code_b_test),
 	    {error, badarg} = code:ensure_loaded(34),
 	    ok;
@@ -379,6 +447,9 @@ delete(Config) when is_list(Config) ->
     ok.
 
 purge(Config) when is_list(Config) ->
+    run_purge_test(fun purge_test/1, Config).
+
+purge_test(Config) when is_list(Config) ->
     OldFlag = process_flag(trap_exit, true),
     code:purge(code_b_test),
     {'EXIT',_} = (catch code:purge({})),
@@ -391,7 +462,13 @@ purge(Config) when is_list(Config) ->
     process_flag(trap_exit, OldFlag),
     ok.
 
+purge_many_exits() ->
+    [{timetrap, {minutes, 2}}].
+
 purge_many_exits(Config) when is_list(Config) ->
+    run_purge_test(fun purge_many_exits_test/1, Config).
+
+purge_many_exits_test(Config) when is_list(Config) ->
     OldFlag = process_flag(trap_exit, true),
 
     code:purge(code_b_test),
@@ -441,6 +518,9 @@ purge_many_exits_do(PurgeF) ->
 
 
 soft_purge(Config) when is_list(Config) ->
+    run_purge_test(fun soft_purge_test/1, Config).
+
+soft_purge_test(Config) when is_list(Config) ->
     OldFlag = process_flag(trap_exit, true),
     code:purge(code_b_test),
     {'EXIT',_} = (catch code:soft_purge(23)),
@@ -498,6 +578,47 @@ all_unique([]) -> ok;
 all_unique([_]) -> ok;
 all_unique([{X,_}|[{Y,_}|_]=T]) when X < Y -> all_unique(T).
 
+all_available(Config) when is_list(Config) ->
+    case test_server:is_cover() of
+	true -> {skip,"Cover is running"};
+	false -> all_available_1(Config)
+    end.
+
+all_available_1(Config) ->
+
+    %% Add an ez dir to make sure the modules in there are found
+    DDir = proplists:get_value(data_dir,Config)++"clash/",
+    true = code:add_path(DDir++"foobar-0.1.ez/foobar-0.1/ebin"),
+
+    Available = code:all_available(),
+
+    %% Test that baz and blarg that are part of the .ez archive are found
+    {value, _} =
+        lists:search(fun({Name,_,Loaded}) -> not Loaded andalso Name =:= "baz" end, Available),
+    {value, _} =
+        lists:search(fun({Name,_,Loaded}) -> not Loaded andalso Name =:= "blarg" end, Available),
+
+    %% Test that all loaded are part of all available
+    Loaded = [{atom_to_list(M),P,true} || {M,P} <- code:all_loaded()],
+    [] = Loaded -- Available,
+
+    {value, {ModStr,_,false} = NotLoaded} =
+        lists:search(fun({Name,_,Loaded}) -> not is_atom(Name) end, Available),
+    ct:log("Testing with ~p",[NotLoaded]),
+
+    Mod = list_to_atom(ModStr),
+
+    %% Test that the module is actually not loaded
+    false = code:is_loaded(Mod),
+
+    %% Load it
+    Mod:module_info(),
+
+    {value, {ModStr,_,true}} =
+        lists:search(fun({Name,_,_}) -> Name =:= ModStr end, code:all_available()),
+
+    ok.
+
 load_binary(Config) when is_list(Config) ->
     TestDir = test_dir(),
     File = TestDir ++ "/code_b_test" ++ code:objfile_extension(),
@@ -514,24 +635,12 @@ load_binary(Config) when is_list(Config) ->
     ok.
 
 
+upgrade() ->
+    [{timetrap,{minutes,2}}].
+
 upgrade(Config) ->
     DataDir = proplists:get_value(data_dir, Config),
-
-    case erlang:system_info(hipe_architecture) of
-        undefined ->
-            upgrade_do(DataDir, beam, [beam]);
-
-        _ ->
-            T = [beam, hipe],
-            [upgrade_do(DataDir, Client, T) || Client <- T],
-
-            case hipe:llvm_support_available() of
-                false -> ok;
-                true  ->
-                    T2 = [beam, hipe_llvm],
-                    [upgrade_do(DataDir, Client, T2) || Client <- T2]
-            end
-    end,
+    upgrade_do(DataDir, beam, [beam]),
     ok.
 
 upgrade_do(DataDir, Client, T) ->
@@ -551,12 +660,8 @@ compile_load(Mod, Dir, Ver, CodeType) ->
 		[Ver, Mod, CodeType]),
 	    [{d,list_to_atom("VERSION_" ++ integer_to_list(Ver))}]
     end,
-    Target = case CodeType of
-	beam -> [];
-	hipe -> [native];
-	hipe_llvm -> [native,{hipe,[to_llvm]}]
-    end,
-    CompOpts = [binary, report] ++ Target ++ Version,
+
+    CompOpts = [binary, report] ++ Version,
 
     Src = filename:join(Dir, atom_to_list(Mod) ++ ".erl"),
     T1 = erlang:now(),
@@ -614,8 +719,7 @@ set_path_file(Config) when is_list(Config) ->
 %% Test that a module with the same name as a module in
 %% a sticky directory cannot be loaded.
 sticky_dir(Config) when is_list(Config) ->
-    Pa = filename:dirname(code:which(?MODULE)),
-    {ok,Node} = test_server:start_node(sticky_dir, slave, [{args,"-pa "++Pa}]),
+    {ok, Peer, Node} = ?CT_PEER(),
     Mods = [code,lists,erlang,init],
     OutDir = filename:join(proplists:get_value(priv_dir, Config), sticky_dir),
     _ = file:make_dir(OutDir),
@@ -628,7 +732,7 @@ sticky_dir(Config) when is_list(Config) ->
 	    io:format("~p\n", [Other]),
 	    ct:fail(failed)
     end,
-    test_server:stop_node(Node),
+    peer:stop(Peer),
     ok.
 
 sticky_compiler(Files, PrivDir) ->
@@ -666,22 +770,17 @@ pa_pz_option(Config) when is_list(Config) ->
     DDir = proplists:get_value(data_dir,Config),
     PaDir = filename:join(DDir,"pa"),
     PzDir = filename:join(DDir,"pz"),
-    {ok, Node}=test_server:start_node(pa_pz1, slave,
-	[{args,
-		"-pa " ++ PaDir
-		++ " -pz " ++ PzDir}]),
+    {ok, Peer, Node} = ?CT_PEER(["-pa", PaDir, "-pz", PzDir]),
     Ret=rpc:call(Node, code, get_path, []),
     [PaDir|Paths] = Ret,
     [PzDir|_] = lists:reverse(Paths),
-    test_server:stop_node(Node),
-    {ok, Node2}=test_server:start_node(pa_pz2, slave,
-	[{args,
-		"-mode embedded " ++ "-pa "
-		++ PaDir ++ " -pz " ++ PzDir}]),
+    peer:stop(Peer),
+    {ok, Peer2, Node2} = ?CT_PEER(["-mode", "embedded", "-pa",
+        PaDir, "-pz", PzDir]),
     Ret2=rpc:call(Node2, code, get_path, []),
     [PaDir|Paths2] = Ret2,
     [PzDir|_] = lists:reverse(Paths2),
-    test_server:stop_node(Node2).
+    peer:stop(Peer2).
 
 %% add_path, del_path should not cause priv_dir(App) to fail.
 add_del_path(Config) when is_list(Config) ->
@@ -733,7 +832,7 @@ clash(Config) when is_list(Config) ->
     true = code:add_path(TmpEzFile++"/foobar-0.1/ebin"),
     case os:type() of
         {win32,_} ->
-	    %% The file wont be deleted on windows until it's closed, why we
+	    %% The file won't be deleted on windows until it's closed, why we
 	    %% need to rename instead.
 	    ok = file:rename(TmpEzFile,TmpEzFile++".moved");
 	 _ ->
@@ -788,8 +887,7 @@ analyse([], [This={M,F,A}|Path], Visited, ErrCnt0) ->
     %% the code_server is started.
     OK = [erlang, os, prim_file, erl_prim_loader, init, ets,
 	  code_server, lists, lists_sort, unicode, binary, filename,
-	  gb_sets, gb_trees, hipe_unified_loader, hipe_bifs,
-	  erts_code_purger,
+	  gb_sets, gb_trees, erts_code_purger, erts_internal, code,
 	  prim_zip, zlib],
     ErrCnt1 =
 	case lists:member(M, OK) or erlang:is_builtin(M,F,A) of
@@ -836,13 +934,15 @@ check_funs({'$M_EXPR','$F_EXPR',_},
 	   [{code_server,do_mod_call,4},
 	    {code_server,handle_call,3}|_]) -> 0;
 check_funs({'$M_EXPR','$F_EXPR',_},
-	   [{lists,flatmap,2},
+	   [{lists,flatmap_1,2},
+	    {lists,flatmap,2},
 	    {lists,concat,1},
 	    {code_server,load_abs,4},
 	    {code_server,handle_call,3},
 	    {code_server,loop,1}|_]) -> 0;
 check_funs({'$M_EXPR','$F_EXPR',_},
-	   [{lists,foreach,2},
+	   [{lists,foreach_1,2},
+	    {lists,foreach,2},
 	    {code_server,stick_dir,3},
 	    {code_server,handle_call,3},
 	    {code_server,loop,1}|_]) -> 0;
@@ -858,6 +958,19 @@ check_funs({'$M_EXPR','$F_EXPR',1},
 	    {code_server,get_user_lib_dirs,0},
 	    {code_server,init,3},
 	    {code_server,start_link,1}]) -> 0;
+check_funs({'$M_EXPR','$F_EXPR',1},
+	   [{lists,all_1,2},
+	    {lists,all,2},
+	    {code_server,is_numstr,1},
+	    {code_server,is_vsn,1},
+	    {code_server,vsn_to_num,1},
+	    {code_server,create_bundle,2},
+	    {code_server,choose_bundles,1},
+	    {code_server,make_path,2},
+	    {code_server,get_user_lib_dirs_1,1},
+	    {code_server,get_user_lib_dirs,0},
+	    {code_server,init,3},
+	    {code_server,start_link,1}]) -> 0;
 check_funs({'$M_EXPR','$F_EXPR',1},
 	   [{lists,filter,2},
 	    {code_server,try_archive_subdirs,3}|_]) -> 0;
@@ -867,18 +980,6 @@ check_funs({'$M_EXPR','$F_EXPR',_},
 	    {code_server,start_link,1}]) -> 0;
 check_funs({'$M_EXPR','$F_EXPR',_},
 	   [{erlang,spawn_link,1},{code_server,start_link,1}]) -> 0;
-check_funs({'$M_EXPR','$F_EXPR',2},
-	   [{hipe_unified_loader,write_words,3} | _]) -> 0;
-check_funs({'$M_EXPR','$F_EXPR',2},
-	   [{hipe_unified_loader,patch_label_or_labels,4} | _]) -> 0;
-check_funs({'$M_EXPR','$F_EXPR',2},
-	   [{hipe_unified_loader,sort_and_write,5} | _]) -> 0;
-check_funs({'$M_EXPR','$F_EXPR',2},
-	   [{lists,foldl,3},
-	    {hipe_unified_loader,sort_and_write,5} | _]) -> 0;
-check_funs({'$M_EXPR','$F_EXPR',1},
-	   [{lists,foreach,2},
-	    {hipe_unified_loader,patch_consts,4} | _]) -> 0;
 check_funs({'$M_EXPR',warning_msg,2},
 	   [{code_server,finish_on_load_report,2} | _]) -> 0;
 check_funs({'$M_EXPR','$F_EXPR',1},
@@ -889,17 +990,6 @@ check_funs({'$M_EXPR','$F_EXPR',2},
 	   [{code_server,handle_pending_on_load,4}|_]) -> 0;
 check_funs({'$M_EXPR','$F_EXPR',2},
 	   [{code_server,finish_on_load_2,3}|_]) -> 0;
-%% This is cheating! /raimo
-%%
-%% check_funs(This = {M,_,_}, Path) ->
-%%     case catch atom_to_list(M) of
-%% 	[$h,$i,$p,$e | _] ->
-%% 	    io:format("hipe_module_ignored(~p, ~p)~n", [This, Path]),
-%% 	    0;
-%% 	_ ->
-%% 	    io:format("not_verified(~p, ~p)~n", [This, Path]),
-%% 	    1
-%%     end;
 check_funs(This, Path) ->
     io:format("not_verified(~p, ~p)~n", [This, Path]),
     1.
@@ -928,6 +1018,9 @@ where_is_file(Config) when is_list(Config) ->
 
 %% Test that stacktrace is deleted when purging a referred module.
 purge_stacktrace(Config) when is_list(Config) ->
+    run_purge_test(fun purge_stacktrace_test/1, Config).
+    
+purge_stacktrace_test(Config) when is_list(Config) ->
     code:purge(code_b_test),
     try code_b_test:call(fun(b) -> ok end, a)
     catch
@@ -936,7 +1029,7 @@ purge_stacktrace(Config) when is_list(Config) ->
 	    case Stacktrace of
 		      [{?MODULE,_,[a],_},
 		       {code_b_test,call,2,_},
-		       {?MODULE,purge_stacktrace,1,_}|_] ->
+		       {?MODULE,purge_stacktrace_test,1,_}|_] ->
 			  false = code:purge(code_b_test)
 		  end
     end,
@@ -946,7 +1039,7 @@ purge_stacktrace(Config) when is_list(Config) ->
 	    code:load_file(code_b_test),
 	    case Stacktrace2 of
 		      [{code_b_test,call,[nofun,2],_},
-		       {?MODULE,purge_stacktrace,1,_}|_] ->
+		       {?MODULE,purge_stacktrace_test,1,_}|_] ->
 			  false = code:purge(code_b_test)
 		  end
     end,
@@ -957,7 +1050,7 @@ purge_stacktrace(Config) when is_list(Config) ->
 	    code:load_file(code_b_test),
 	    case Stacktrace3 of
 		      [{code_b_test,call,Args,_},
-		       {?MODULE,purge_stacktrace,1,_}|_] ->
+		       {?MODULE,purge_stacktrace_test,1,_}|_] ->
 			  false = code:purge(code_b_test)
 		  end
     end,
@@ -966,19 +1059,23 @@ purge_stacktrace(Config) when is_list(Config) ->
 mult_lib_roots(Config) when is_list(Config) ->
     DataDir = filename:join(proplists:get_value(data_dir, Config), "mult_lib_roots"),
     mult_lib_compile(DataDir, "my_dummy_app-b/ebin/lists"),
-    mult_lib_compile(DataDir,
-			   "my_dummy_app-c/ebin/code_SUITE_mult_root_module"),
 
-    %% Set up ERL_LIBS and start a slave node.
+    %% We will use this module to test both code path loading and
+    %% code path caching. So we ensure its beam file does not exist.
+    CodeSuiteMultRootBeam =
+	filename:join(DataDir, "first_root/my_dummy_app-c/ebin/code_SUITE_mult_root_module.beam"),
+    file:delete(CodeSuiteMultRootBeam),
+
+    %% Set up ERL_LIBS and start a peer node.
     ErlLibs = filename:join(DataDir, "first_root") ++ mult_lib_sep() ++
 	filename:join(DataDir, "second_root"),
 
-    {ok,Node} =
-	test_server:start_node(mult_lib_roots, slave,
-		      [{args,"-env ERL_LIBS "++ErlLibs}]),
+    {ok, Peer, Node} = ?CT_PEER(["-env", "ERL_LIBS", ErlLibs]),
 
     Path0 = rpc:call(Node, code, get_path, []),
-    ["."|Path1] = Path0,
+    %% ?CT_PEER adds extra path to this module folder
+    PathToSelf = filename:dirname(code:which(?MODULE)),
+    [PathToSelf, "."|Path1] = Path0,
     [Kernel|Path2] = Path1,
     [Stdlib|Path3] = Path2,
     mult_lib_verify_lib(Kernel, "kernel"),
@@ -995,8 +1092,31 @@ mult_lib_roots(Config) when is_list(Config) ->
 	    E <- lists:sort([Lib1,Lib2,Lib3,Lib4,Lib5])],
     io:format("~p\n", [Path]),
 
+    %% Now let's attempt to dynamically add a module,
+    %% this will fail due to the boot paths cache.
+    error = rpc:call(Node, code, get_object_code, [code_SUITE_mult_root_module]),
+    mult_lib_compile(DataDir, "my_dummy_app-c/ebin/code_SUITE_mult_root_module"),
+    error = rpc:call(Node, code, get_object_code, [code_SUITE_mult_root_module]),
+    %% Test that the CWD is not cached
+    File = filename:join([DataDir, "first_root/my_dummy_app-c/ebin/code_SUITE_mult_root_module"]),
+    {ok, _} = compile:file(File, [{outdir, "."}]),
+    {_,_,_} = rpc:call(Node, code, get_object_code, [code_SUITE_mult_root_module]),
     true = rpc:call(Node, code_SUITE_mult_root_module, works_fine, []),
+    file:delete("code_SUITE_mult_root_module.beam"),
+    %% Clean up so we can start again
+    file:delete(CodeSuiteMultRootBeam),
+    peer:stop(Peer),
 
+    NoCacheArgs = ["-env", "ERL_LIBS", ErlLibs, "-cache_boot_paths", "false"],
+    {ok, NoCachePeer, NoCacheNode} = ?CT_PEER(NoCacheArgs),
+
+    %% Now the same code should work
+    error = rpc:call(NoCacheNode, code, get_object_code, [code_SUITE_mult_root_module]),
+    mult_lib_compile(DataDir, "my_dummy_app-c/ebin/code_SUITE_mult_root_module"),
+    {_,_,_} = rpc:call(NoCacheNode, code, get_object_code, [code_SUITE_mult_root_module]),
+    true = rpc:call(NoCacheNode, code_SUITE_mult_root_module, works_fine, []),
+
+    peer:stop(NoCachePeer),
     ok.
 
 mult_lib_compile(Root, Last) ->
@@ -1021,17 +1141,20 @@ mult_lib_remove_prefix([H|T1], [H|T2]) ->
 mult_lib_remove_prefix([$/|T], []) -> T.
 
 bad_erl_libs(Config) when is_list(Config) ->
-    {ok,Node} =
-	test_server:start_node(bad_erl_libs, slave, []),
+    %% Preserve ERL_LIBS if set.
+    BadLibs0 = "/no/such/dir",
+    BadLibs =
+         case os:getenv("ERL_LIBS") of
+             false -> BadLibs0;
+             Libs -> BadLibs0 ++ ":" ++ Libs
+         end,
+    {ok, Peer, Node} = ?CT_PEER(),
     Code = rpc:call(Node,code,get_path,[]),
-    test_server:stop_node(Node),
-
-    {ok,Node2} =
-	test_server:start_node(bad_erl_libs, slave,
-			       [{args,"-env ERL_LIBS /no/such/dir"}]),
-    Code2 = rpc:call(Node,code,get_path,[]),
-    test_server:stop_node(Node2),
+    peer:stop(Peer),
 
+    {ok, Peer2, Node2} = ?CT_PEER(["-env", "ERL_LIBS", BadLibs]),
+    Code2 = rpc:call(Node2,code,get_path,[]),
+    peer:stop(Peer2),
     %% Test that code path is not affected by the faulty ERL_LIBS
     Code = Code2,
 
@@ -1089,10 +1212,8 @@ do_code_archive(Config, Root, StripVsn) when is_list(Config) ->
     filelib:ensure_dir(OtherFile),
     ok = file:write_file(OtherFile, OtherContents),
 
-    %% Set up ERL_LIBS and start a slave node.
-    {ok, Node} =
-	test_server:start_node(code_archive, slave,
-		      [{args,"-env ERL_LIBS " ++ RootDir}]),
+    %% Set up ERL_LIBS and start a peer node.
+    {ok, Peer, Node} = ?CT_PEER(["-env", "ERL_LIBS", RootDir]),
     CodePath = rpc:call(Node, code, get_path, []),
     AppEbin = filename:join([Archive, Base, "ebin"]),
     io:format("AppEbin: ~p\n", [AppEbin]),
@@ -1134,7 +1255,7 @@ do_code_archive(Config, Root, StripVsn) when is_list(Config) ->
     error =  rpc:call(Node, App, find, [Tab, Key]),
     ok =  rpc:call(Node, App, erase, [Tab]),
 
-    test_server:stop_node(Node),
+    peer:stop(Peer),
     ok.
 
 compile_app(TopDir, AppName) ->
@@ -1164,12 +1285,11 @@ compile_files([], _, _) ->
 %% embeddedd system.
 big_boot_embedded(Config) when is_list(Config) ->
     {BootArg,AppsInBoot} = create_big_boot(Config),
-    {ok, Node} =
-	test_server:start_node(big_boot_embedded, slave,
-		      [{args,"-boot "++BootArg++" -mode embedded"}]),
+    {ok, Peer, Node} = ?CT_PEER(["-boot", BootArg, "-mode", "embedded"]),
     RemoteNodeApps =
 	[ {X,Y} || {X,_,Y} <-
 		       rpc:call(Node,application,loaded_applications,[]) ],
+    peer:stop(Peer),
     true = lists:sort(AppsInBoot) =:=  lists:sort(RemoteNodeApps),
     ok.
 
@@ -1315,12 +1435,11 @@ on_load_embedded_1(Config) ->
     true = code:del_path(OnLoadAppEbin),
 
     %% Start the node and check that the on_load function was run.
-    {ok,Node} = start_node(on_load_embedded,
-				 "-mode embedded -boot " ++ BootScript),
+    {ok, Peer, Node} = ?CT_PEER(["-mode", "embedded", "-boot", BootScript]),
     ok = rpc:call(Node, on_load_embedded, status, []),
 
     %% Clean up.
-    stop_node(Node).
+    peer:stop(Peer).
 
 del_link(LinkName) ->
    case file:delete(LinkName) of
@@ -1367,12 +1486,6 @@ create_big_boot(Config) ->
     {filename:join(LatestDir, LatestName),Apps}.
 
 %% The following apps cannot be loaded.
-%% hipe .app references (or can reference) files that have no
-%% corresponding beam file (if hipe is not enabled).
-filter_app("hipe",_) -> false;
-
-%% Dialyzer depends on hipe
-filter_app("dialyzer",_) -> false;
 
 %% Orber requires explicit configuration
 filter_app("orber",_) -> false;
@@ -1400,6 +1513,9 @@ filter_app("os_mon",true) -> false;
 %% erts is not a "real" app either =/
 filter_app("erts",_) -> false;
 
+%% wx* depends on that wxwidgets libs
+filter_app("wx"++_,_) -> false;
+
 %% Other apps should be OK.
 filter_app(_,_) -> true.
 
@@ -1550,6 +1666,11 @@ on_load_update_code_1(3, Mod) ->
 
 %% Test -on_load while trace feature 'on_load' is enabled (OTP-14612)
 on_load_trace_on_load(Config) ->
+    %% 'on_load' enables tracing for all newly loaded modules, so we make a dry
+    %% run to ensure that ancillary modules like 'merl' won't be loaded during
+    %% the actual test.
+    on_load_update(Config),
+
     Papa = self(),
     Tracer = spawn_link(fun F() -> receive M -> Papa ! M end, F() end),
     {tracer,[]} = erlang:trace_info(self(),tracer),
@@ -1572,8 +1693,10 @@ flush() ->
     after 100 -> []
     end.
 
+on_load_purge(Config) when is_list(Config) ->
+    run_purge_test(fun on_load_purge_test/1, Config).
 
-on_load_purge(_Config) ->
+on_load_purge_test(Config) when is_list(Config) ->
     Mod = ?FUNCTION_NAME,
     register(Mod, self()),
     Tree = ?Q(["-module('@Mod@').\n",
@@ -1602,7 +1725,9 @@ on_load_purge(_Config) ->
 	    after 10000 ->
 		    ct:fail(no_down_message)
 	    end
-    end.
+    end,
+    unregister(Mod),
+    ok.
 
 on_load_self_call(_Config) ->
     Mod = ?FUNCTION_NAME,
@@ -1623,9 +1748,9 @@ on_load_self_call(_Config) ->
     ok.
 
 on_load_do_load(Mod, Code) ->
-    spawn(fun() ->
-		  {module,Mod} = code:load_binary(Mod, "", Code)
-	  end),
+    spawn_link(fun() ->
+                       {module,Mod} = code:load_binary(Mod, "", Code)
+               end),
     receive
 	Any -> Any
     end.
@@ -1716,9 +1841,9 @@ on_load_deleted(_Config) ->
 			    "  receive _ -> ok end.\n"]),
 		 merl:print(Tree),
 		 {ok,Mod,Code} = merl:compile(Tree),
-		 spawn(fun() ->
-			       {module,Mod} = code:load_binary(Mod, "", Code)
-		       end),
+		 spawn_link(fun() ->
+                                    {module,Mod} = code:load_binary(Mod, "", Code)
+                            end),
 		 receive after 1 -> ok end,
 		 {module,OtherMod} = code:load_binary(OtherMod, "",
 						      OtherCode),
@@ -1741,10 +1866,10 @@ delete_before_reload(Mod, Reload) ->
     {ok,Mod,Code1} = merl:compile(Tree1),
 
     Self = self(),
-    spawn(fun() ->
-		  {module,Mod} = code:load_binary(Mod, "", Code1),
-		  Mod:f(Self)
-	  end),
+    spawn_link(fun() ->
+                       {module,Mod} = code:load_binary(Mod, "", Code1),
+                       Mod:f(Self)
+               end),
     receive started -> ok end,
 
     true = code:delete(Mod),
@@ -1769,35 +1894,6 @@ delete_before_reload(Mod, Reload) ->
 
     ok.
 
-
-%% Test that the native code of early loaded modules is loaded.
-native_early_modules(Config) when is_list(Config) ->
-    case erlang:system_info(hipe_architecture) of
-	undefined ->
-	    {skip,"Native code support is not enabled"};
-	Architecture ->
-	    native_early_modules_1(Architecture)
-    end.
-
-native_early_modules_1(Architecture) ->
-    {lists, ListsBinary, _ListsFilename} = code:get_object_code(lists),
-    ChunkName = hipe_unified_loader:chunk_name(Architecture),
-    NativeChunk = beam_lib:chunks(ListsBinary, [ChunkName]),
-    IsHipeCompiled = case NativeChunk of
-        {ok,{_,[{_,Bin}]}} when is_binary(Bin) -> true;
-        {error, beam_lib, _} -> false
-    end,
-    case IsHipeCompiled of
-        false ->
-	    {skip,"OTP apparently not configured with --enable-native-libs"};
-        true ->
-	    true = lists:all(fun code:is_module_native/1,
-		[ets,file,filename,gb_sets,gb_trees,
-		    %%hipe_unified_loader, no_native as workaround
-		    lists,os]),
-            ok
-    end.
-
 %% Test that the mode of the code server is properly retrieved.
 get_mode(Config) when is_list(Config) ->
     interactive = code:get_mode().
@@ -1817,6 +1913,28 @@ do_normalized_paths([M|Ms]) ->
 do_normalized_paths([]) ->
     ok.
 
+%% Make sure that the extra -mode flags are ignored
+mult_embedded_flags(_Config) ->
+    Modes = [{["-mode", "embedded"], embedded},
+	     {["-mode", "interactive"], interactive},
+	     {["-mode", "invalid"], interactive}],
+
+    [ begin
+	  {ArgMode, ExpectedMode} = Mode,
+	  {ok, Peer, Node} = ?CT_PEER(ArgMode),
+	  ExpectedMode = rpc:call(Node, code, get_mode, []),
+	  peer:stop(Peer)
+      end || Mode <- Modes],
+
+    [ begin
+	  {ArgIgnoredMode, _} = IgnoredMode,
+	  {ArgRelevantMode, ExpectedMode} = RelevantMode,
+	  {ok, Peer, Node} = ?CT_PEER(ArgRelevantMode ++ ArgIgnoredMode),
+	  ExpectedMode = rpc:call(Node, code, get_mode, []),
+	  peer:stop(Peer)
+      end || IgnoredMode <- Modes, RelevantMode <- Modes],
+    ok.
+
 %% Test that module_status/1 behaves as expected
 module_status(_Config) ->
     case test_server:is_cover() of
@@ -1838,6 +1956,11 @@ module_status() ->
     loaded = code:module_status(erlang),        % preloaded
     loaded = code:module_status(?MODULE),       % normal known loaded
 
+    %% module_status/0 returns status for each loaded module
+    true = (lists:sort([{M, code:module_status(M)}
+                        || {M, _} <- code:all_loaded()])
+            =:= lists:sort(code:module_status())),
+
     non_existing = code:which(?TESTMOD), % verify dummy name not in path
     code:purge(?TESTMOD), % ensure no previous version in memory
     code:delete(?TESTMOD),
@@ -1850,6 +1973,11 @@ module_status() ->
     "" = code:which(?TESTMOD), % verify empty string for source file
     loaded = code:module_status(?TESTMOD),
 
+    %% module_status/1 also accepts a list of modules
+    [] = code:module_status([]),
+    [{erlang, loaded},{?MODULE,loaded},{?TESTMOD,loaded}] =
+        code:module_status([erlang, ?MODULE, ?TESTMOD]),
+
     %% deleting generated code
     true = code:delete(?TESTMOD),
     non_existing = code:which(?TESTMOD), % verify still not in path
@@ -1899,64 +2027,11 @@ module_status() ->
     compile_beam(0),
     loaded = code:module_status(?TESTMOD),
 
-    case erlang:system_info(hipe_architecture) of
-	undefined ->
-	    %% no native support
-	    ok;
-	_ ->
-	    %% native chunk is ignored if beam code is already loaded
-	    load_code(),
-	    loaded = code:module_status(?TESTMOD),
-	    false = has_native(?TESTMOD),
-	    compile_native(0),
-	    BeamMD5 = erlang:get_module_info(?TESTMOD, md5),
-	    {ok,{?TESTMOD,BeamMD5}} = beam_lib:md5(?TESTMODOBJ), % beam md5 unchanged
-	    loaded = code:module_status(?TESTMOD),
-
-	    %% native code reported as loaded, though different md5 from beam
-	    load_code(),
-	    true = has_native(?TESTMOD),
-	    NativeMD5 = erlang:get_module_info(?TESTMOD, md5),
-	    true = (BeamMD5 =/= NativeMD5),
-	    loaded = code:module_status(?TESTMOD),
-
-	    %% recompilation ignores timestamps, only md5 matters
-	    compile_native(1100), % later timestamp
-	    loaded = code:module_status(?TESTMOD),
-
-	    %% modifying native module detects different md5
-	    make_source_file(<<"2">>),
-	    compile_native(0),
-	    modified = code:module_status(?TESTMOD),
-
-	    %% loading the modified native code from disk makes it loaded
-	    load_code(),
-	    true = has_native(?TESTMOD),
-	    NativeMD5_2 = erlang:get_module_info(?TESTMOD, md5),
-	    true = (NativeMD5 =/= NativeMD5_2), % verify native md5 changed
-	    {ok,{?TESTMOD,BeamMD5_2}} = beam_lib:md5(?TESTMODOBJ),
-	    true = (BeamMD5_2 =/= NativeMD5_2), % verify md5 differs from beam
-	    loaded = code:module_status(?TESTMOD),
-
-	    %% removing and recreating a native module with same md5
-	    remove_code(),
-	    removed = code:module_status(?TESTMOD),
-	    compile_native(0),
-	    loaded = code:module_status(?TESTMOD),
-
-	    %% purging/deleting native module
-	    code:purge(?TESTMOD),
-	    true = code:delete(?TESTMOD),
-	    not_loaded = code:module_status(?TESTMOD)
-    end,
     ok.
 
 compile_beam(Sleep) ->
     compile(Sleep, []).
 
-compile_native(Sleep) ->
-    compile(Sleep, [native]).
-
 compile(Sleep, Opts) ->
     timer:sleep(Sleep),  % increment compilation timestamp
     {ok,?TESTMOD} = compile:file(?TESTMODSRC, Opts).
@@ -1968,12 +2043,6 @@ load_code() ->
 remove_code() ->
     ok = file:delete(?TESTMODOBJ).
 
-has_native(Module) ->
-    case erlang:get_module_info(Module, native_addresses) of
-	[] -> false;
-	[_|_] -> true
-    end.
-
 make_source_file(Body) ->
     ok = file:write_file(?TESTMODSRC, dummy_source(Body)).
 
@@ -2022,8 +2091,38 @@ terminate(_Reason, State) ->
 %%% Common utility functions.
 %%%
 
-start_node(Name, Param) ->
-    test_server:start_node(Name, slave, [{args, Param}]).
+run_purge_test(Test, Config) ->
+    OSRL = erlang:system_info(outstanding_system_requests_limit),
+    
+    TestLowOSRL = case OSRL < 10 of
+                      true -> 1;
+                      false -> 5
+                  end,
 
-stop_node(Node) ->
-    test_server:stop_node(Node).
+    io:format("Running with the default outstanding request limit of ~p~n", [OSRL]),
+    Res1 = Test(Config),
+    io:format("Result: ~p~n", [Res1]),
+    try
+        %% Run again with low limit and many processes...
+        Procs = case erlang:system_info(process_limit) of
+                    PLim when PLim < 20000 ->
+                        erlang:system_info(process_limit) div 2;
+                    _ ->
+                        10000
+                end,
+        erlang:system_flag(outstanding_system_requests_limit, TestLowOSRL),
+        io:format("Running with outstanding request limit of ~p~n", [TestLowOSRL]),
+        Ps = lists:map(fun (_) ->
+                               spawn_link(fun () -> receive after infinity -> ok end end)
+                       end, lists:seq(1, Procs)),
+        Res2 = Test(Config),
+        lists:foreach(fun (P) ->
+                              unlink(P),
+                              exit(P, kill)
+                      end, Ps),
+        lists:foreach(fun (P) -> is_process_alive(P) end, Ps),
+        io:format("Result: ~p~n", [Res2]),
+        {Res1, Res2}
+    after
+        TestLowOSRL = erlang:system_flag(outstanding_system_requests_limit, OSRL)
+    end.
