@@ -1,7 +1,7 @@
 %%
 %% %CopyrightBegin%
 %%
-%% Copyright Ericsson AB 1997-2021. All Rights Reserved.
+%% Copyright Ericsson AB 1997-2023. All Rights Reserved.
 %%
 %% Licensed under the Apache License, Version 2.0 (the "License");
 %% you may not use this file except in compliance with the License.
@@ -33,7 +33,7 @@
 
 	 t_gethostbyaddr/0, t_gethostbyaddr/1,
 	 t_getaddr/0, t_getaddr/1,
-	 t_gethostbyname/0, t_gethostbyname/1,
+	 t_gethostbyname/0, t_gethostbyname/1, t_gethostbyname_empty/1,
 	 t_gethostbyaddr_v6/0, t_gethostbyaddr_v6/1,
 	 t_getaddr_v6/0, t_getaddr_v6/1,
 	 t_gethostbyname_v6/0, t_gethostbyname_v6/1,
@@ -47,10 +47,12 @@
 	 lookup_bad_search_option/1,
 	 getif/1,
 	 getif_ifr_name_overflow/1,getservbyname_overflow/1, getifaddrs/1,
-	 parse_strict_address/1, ipv4_mapped_ipv6_address/1,
+	 is_ip_address/1,
+	 parse_strict_address/1, ipv4_mapped_ipv6_address/1, ntoa/1,
          simple_netns/1, simple_netns_open/1,
          add_del_host/1, add_del_host_v6/1,
-         simple_bind_to_device/1, simple_bind_to_device_open/1
+         simple_bind_to_device/1, simple_bind_to_device_open/1,
+	 socknames_sctp/1, socknames_tcp/1, socknames_udp/1
         ]).
 
 -export([
@@ -68,22 +70,25 @@ suite() ->
 
 all() -> 
     [
-     t_gethostbyaddr, t_gethostbyname, t_getaddr,
+     t_gethostbyaddr, t_gethostbyname, t_gethostbyname_empty, t_getaddr,
      t_gethostbyaddr_v6, t_gethostbyname_v6, t_getaddr_v6,
-     ipv4_to_ipv6, host_and_addr, {group, parse},
+     ipv4_to_ipv6, host_and_addr, is_ip_address, {group, parse},
      t_gethostnative, gethostnative_parallell, cname_loop,
      missing_hosts_reload, hosts_file_quirks,
      gethostnative_debug_level, gethostnative_soft_restart,
      lookup_bad_search_option,
      getif, getif_ifr_name_overflow, getservbyname_overflow,
-     getifaddrs, parse_strict_address, simple_netns, simple_netns_open,
+     getifaddrs, parse_strict_address, ipv4_mapped_ipv6_address, ntoa,
+     simple_netns, simple_netns_open,
      add_del_host, add_del_host_v6,
-     simple_bind_to_device, simple_bind_to_device_open
+     simple_bind_to_device, simple_bind_to_device_open,
+     {group, socknames}
     ].
 
 groups() -> 
     [
-     {parse, [], parse_cases()}
+     {parse,     [], parse_cases()},
+     {socknames, [], socknames_cases()}
     ].
 
 parse_cases() ->
@@ -92,7 +97,14 @@ parse_cases() ->
      parse_address
     ].
 
-%% Required configuaration
+socknames_cases() ->
+    [
+     socknames_sctp,
+     socknames_tcp,
+     socknames_udp
+    ].
+
+%% Required configuration
 required(v4) ->
     [{require, test_host_ipv4_only},
      {require, test_dummy_host}];
@@ -216,19 +228,38 @@ end_per_testcase2(lookup_bad_search_option, Config) ->
 end_per_testcase2(_Func, _Config) ->
     ok.
 
-t_gethostbyaddr() ->
-    required(v4).
+t_gethostbyaddr() -> required(v4).
 %% Test the inet:gethostbyaddr/1 function.
 t_gethostbyaddr(Config) when is_list(Config) ->
-    {Name,FullName,IPStr,{A,B,C,D}=IP,Aliases,_,_} = ct:get_config(test_host_ipv4_only),
+    ?TC_TRY(?FUNCTION_NAME, fun() -> do_gethostbyaddr(Config) end).
+
+do_gethostbyaddr(Config) when is_list(Config) ->
+    ?P("begin - try get config 'test_host_ipv4_only'"),
+    {Name,FullName,IPStr,{A,B,C,D}=IP,Aliases,_,_} =
+        ct:get_config(test_host_ipv4_only),
+    ?P("config 'test_host_ipv4_only': "
+       "~n   Name:      ~p"
+       "~n   Full Name: ~p"
+       "~n   IPStr:     ~p"
+       "~n   (IP) A:    ~p"
+       "~n   (IP) B:    ~p"
+       "~n   (IP) C:    ~p"
+       "~n   (IP) D:    ~p"
+       "~n   Aliases:   ~p",
+       [Name, FullName, IPStr, A, B, C, D,Aliases]),
     Rname = integer_to_list(D) ++ "." ++
 	integer_to_list(C) ++ "." ++
 	integer_to_list(B) ++ "." ++
 	integer_to_list(A) ++ ".in-addr.arpa",
-    {ok,HEnt} = inet:gethostbyaddr(IPStr),
-    {ok,HEnt} = inet:gethostbyaddr(IP),
-    {error,Error} = inet:gethostbyaddr(Name),
-    ok = io:format("Failure reason: ~p: ~s", [error,inet:format_error(Error)]),
+    {ok, HEnt} = inet:gethostbyaddr(IPStr),
+    {ok, HEnt} = inet:gethostbyaddr(IP),
+    ?P("gethostbyaddr for (both):"
+       "~n   IPStr: ~p"
+       "~n   IP:    ~p"
+       "~n   => ~p", [IPStr, IP, HEnt]),
+    {error, Reason} = inet:gethostbyaddr(Name),
+    ok = ?P("Expected error with failure reason: "
+            "~n   (~w) ~s", [Reason, inet:format_error(Reason)]),
     HEnt_ = HEnt#hostent{h_addrtype = inet,
                          h_length = 4,
                          h_addr_list = [IP]},
@@ -237,27 +268,34 @@ t_gethostbyaddr(Config) when is_list(Config) ->
         {{unix,freebsd},{5,0,0}} ->
             %% The alias list seems to be buggy in FreeBSD 5.0.0.
             check_elems([{HEnt#hostent.h_name,[Name,FullName]}]),
-            io:format("Buggy alias list: ~p", [HEnt#hostent.h_aliases]),
+            ?P("Buggy alias list: ~p", [HEnt#hostent.h_aliases]),
             ok;
         _ ->
-            io:format("alias list: ~p", [HEnt#hostent.h_aliases]),
-            io:format(
-	      "check alias list: ~p", [[Aliases,tl(Aliases),[Rname]]]),
-            io:format("name: ~p", [HEnt#hostent.h_name]),
-            io:format("check name: ~p", [[Name,FullName]]),
+            ?P("alias list: "
+               "~n   ~p", [HEnt#hostent.h_aliases]),
+            ?P("check alias list: "
+               "~n   ~p", [[Aliases,tl(Aliases),[Rname]]]),
+            ?P("name:       ~p", [HEnt#hostent.h_name]),
+            ?P("check name: ~p", [[Name,FullName]]),
             check_elems(
 	      [{HEnt#hostent.h_name,[Name,FullName]},
 	       {HEnt#hostent.h_aliases,[[],Aliases,tl(Aliases),[Rname]]}])
     end,
 
+    ?P("try get config 'test_dummy_host'"),
     {_DName, _DFullName, DIPStr, DIP, _, _, _} = ct:get_config(test_dummy_host),
     {error,nxdomain} = inet:gethostbyaddr(DIPStr),
     {error,nxdomain} = inet:gethostbyaddr(DIP),
+
+    ?P("done"),
     ok.
 
 t_gethostbyaddr_v6() -> required(v6).
 %% Test the inet:gethostbyaddr/1 inet6 function.
 t_gethostbyaddr_v6(Config) when is_list(Config) ->
+    ?TC_TRY(?FUNCTION_NAME, fun() -> do_gethostbyaddr_v6(Config) end).
+
+do_gethostbyaddr_v6(Config) when is_list(Config) ->
     {Name6, FullName6, IPStr6, IP6, Aliases6} =
 	ct:get_config(test_host_ipv6_only),
 
@@ -266,11 +304,11 @@ t_gethostbyaddr_v6(Config) when is_list(Config) ->
 	%% looking up the host. DNS lookup will probably fail.
 	{error,nxdomain} ->
 	    {skip, "IPv6 test fails! IPv6 not supported on this host!?"};
-	{ok,HEnt6} ->
-	    {ok,HEnt6} = inet:gethostbyaddr(IP6),
-	    {error,Error6} = inet:gethostbyaddr(Name6),
-	    ok = io:format("Failure reason: ~p: ~s",
-			   [Error6, inet:format_error(Error6)]),
+	{ok, HEnt6} ->
+	    {ok, HEnt6} = inet:gethostbyaddr(IP6),
+	    {error, Reason6} = inet:gethostbyaddr(Name6),
+	    ok = ?P("Expected error with failure reason: "
+                    "~n   (~w) ~s", [Reason6, inet:format_error(Reason6)]),
 	    HEnt6_ = HEnt6#hostent{h_addrtype = inet6,
 				   h_length = 16,
 				   h_addr_list = [IP6]},
@@ -289,6 +327,9 @@ t_gethostbyaddr_v6(Config) when is_list(Config) ->
 t_gethostbyname() -> required(v4).
 %% Test the inet:gethostbyname/1 function.
 t_gethostbyname(Config) when is_list(Config) ->
+    ?TC_TRY(?FUNCTION_NAME, fun() -> do_gethostbyname(Config) end).
+
+do_gethostbyname(Config) when is_list(Config) ->
     {Name,FullName,IPStr,IP,Aliases,IP_46_Str,_} =
 	ct:get_config(test_host_ipv4_only),
     {ok,_} = inet:gethostbyname(IPStr),
@@ -326,9 +367,24 @@ t_gethostbyname(Config) when is_list(Config) ->
     {error,nxdomain} = inet:gethostbyname(IP_46_Str),
     ok.
 
+
+t_gethostbyname_empty(Config) when is_list(Config) ->
+    element(1, os:type()) =:= unix andalso
+        begin
+            {error,nxdomain} = inet:gethostbyname(""),
+            {error,nxdomain} = inet:gethostbyname('')
+        end,
+    {error,nxdomain} = inet:gethostbyname("."),
+    {error,nxdomain} = inet:gethostbyname('.'),
+    ok.
+
+
 t_gethostbyname_v6() -> required(v6).
 %% Test the inet:gethostbyname/1 inet6 function.
 t_gethostbyname_v6(Config) when is_list(Config) ->
+    ?TC_TRY(?FUNCTION_NAME, fun() -> do_gethostbyname_v6(Config) end).
+
+do_gethostbyname_v6(Config) when is_list(Config) ->
     {Name, FullName, IPStr, IP, Aliases} =
 	ct:get_config(test_host_ipv6_only),
 
@@ -379,6 +435,9 @@ t_gethostbyname_v6(Config) when is_list(Config) ->
     end.
 
 check_elems([{Val,Tests} | Elems]) ->
+    ?P("check_elems -> entry with"
+       "~n   Val:   ~p"
+       "~n   Tests: ~p", [Val, Tests]),
     check_elem(Val, Tests, Tests),
     check_elems(Elems);
 check_elems([]) -> ok.
@@ -475,7 +534,7 @@ ipv4_to_ipv6(Config) when is_list(Config) ->
 
 
 %% Test looking up hosts and addresses. Use 'ypcat hosts'
-%% or the local eqivalent to find all hosts.
+%% or the local equivalent to find all hosts.
 
 host_and_addr() ->
     ?P("host_and_addr -> entry"),
@@ -749,18 +808,18 @@ parse_address(Config) when is_list(Config) ->
 t_parse_address(Func, _Reversable, []) ->
     io:format("~p done.~n", [Func]),
     ok;
-t_parse_address(Func, Reversable, [{Addr,String}|L]) ->
+t_parse_address(Func, Reversible, [{Addr,String}|L]) ->
     io:format("~p = ~p.~n", [Addr,String]),
     {ok,Addr} = inet:Func(String),
-    case Reversable of
+    case Reversible of
         true ->String = inet:ntoa(Addr);
         false -> ok
     end,
-    t_parse_address(Func, Reversable, L);
-t_parse_address(Func, Reversable, [String|L]) ->
+    t_parse_address(Func, Reversible, L);
+t_parse_address(Func, Reversible, [String|L]) ->
     io:format("~p.~n", [String]),
     {error,einval} = inet:Func(String),
-    t_parse_address(Func, Reversable, L).
+    t_parse_address(Func, Reversible, L).
 
 parse_strict_address(Config) when is_list(Config) ->
     {ok, {127,0,0,1}} =
@@ -770,6 +829,63 @@ parse_strict_address(Config) when is_list(Config) ->
     {ok, {3089,3106,23603,50240,0,0,119,136}} =
 	inet:parse_strict_address("c11:0c22:5c33:c440::077:0088").
 
+is_ip_address(Config) when is_list(Config) ->
+    IPv4Addresses = [
+        {0, 0, 0, 0},
+        {255, 255, 255, 255}
+    ],
+    IPv6Addresses = [
+        {0, 0, 0, 0, 0, 0, 0, 0},
+        {16#ffff, 16#ffff, 16#ffff, 16#ffff, 16#ffff, 16#ffff, 16#ffff, 16#ffff}
+    ],
+    NonIPAddresses = [
+        foo,
+        "0.0.0.0",
+        {},
+        {0},
+        {0, 0},
+        {0, 0, 0},
+        {0, 0, 0, 0, 0},
+        {0, 0, 0, 0, 0, 0},
+        {0, 0, 0, 0, 0, 0, 0},
+        {0, 0, 0, 0, 0, 0, 0, 0, 0},
+        {0, 0, 0, foo},
+        {0, 0, 0, 0, 0, 0, 0, foo},
+        {0, 0, 0, 256},
+        {0, 0, 256, 0},
+        {0, 256, 0, 0},
+        {256, 0, 0, 0},
+        {0, 0, 0, 0, 0, 0, 0, 16#10000},
+        {0, 0, 0, 0, 0, 0, 16#10000, 0},
+        {0, 0, 0, 0, 0, 16#10000, 0, 0},
+        {0, 0, 0, 0, 16#10000, 0, 0, 0},
+        {0, 0, 0, 16#10000, 0, 0, 0, 0},
+        {0, 0, 16#10000, 0, 0, 0, 0, 0},
+        {0, 16#10000, 0, 0, 0, 0, 0, 0},
+        {16#10000, 0, 0, 0, 0, 0, 0, 0},
+        {0, 0, 0, -1},
+        {0, 0, -1, 0},
+        {0, -1, 0, 0},
+        {-1, 0, 0, 0},
+        {0, 0, 0, 0, 0, 0, 0, -1},
+        {0, 0, 0, 0, 0, 0, -1, 0},
+        {0, 0, 0, 0, 0, -1, 0, 0},
+        {0, 0, 0, 0, -1, 0, 0, 0},
+        {0, 0, 0, -1, 0, 0, 0, 0},
+        {0, 0, -1, 0, 0, 0, 0, 0},
+        {0, -1, 0, 0, 0, 0, 0, 0},
+        {-1, 0, 0, 0, 0, 0, 0, 0}
+    ],
+
+    true = lists:all(fun inet:is_ipv4_address/1, IPv4Addresses),
+    false = lists:any(fun inet:is_ipv4_address/1, IPv6Addresses ++ NonIPAddresses),
+
+    true = lists:all(fun inet:is_ipv6_address/1, IPv6Addresses),
+    false = lists:any(fun inet:is_ipv6_address/1, IPv4Addresses ++ NonIPAddresses),
+
+    true = lists:all(fun inet:is_ip_address/1, IPv6Addresses ++ IPv4Addresses),
+    false = lists:any(fun inet:is_ip_address/1, NonIPAddresses).
+
 ipv4_mapped_ipv6_address(Config) when is_list(Config) ->
     {D1,D2,D3,D4} = IPv4Address =
         {rand:uniform(256) - 1,
@@ -790,6 +906,51 @@ ipv4_mapped_ipv6_address(Config) when is_list(Config) ->
     IPv4Address = inet:ipv4_mapped_ipv6_address(IPv6Address),
     ok.
 
+
+ntoa(Config) when is_list(Config) ->
+    M8 = 1 bsl 8,
+    M16 = 1 bsl 16,
+    V4Xs = rand_tuple(4, M8),
+    V6Xs = rand_tuple(4, M16),
+    ntoa(
+      [{A, B, C, D} ||
+          A <- [0, element(1, V4Xs), M8-1, -1, 256],
+          B <- [0, element(2, V4Xs), M8-1, -1, 256],
+          C <- [0, element(3, V4Xs), M8-1, -1, 256],
+          D <- [0, element(4, V4Xs), M8-1, -1, 256]], M8-1),
+    ntoa(
+      [{E, F, G, H, G, G, E, F} ||
+          E <- [0, element(1, V6Xs), M16-1, -1, M16],
+          F <- [0, element(2, V6Xs), M16-1, -1, M16],
+          G <- [0, element(3, V6Xs), M16-1, -1, M16],
+          H <- [0, element(4, V6Xs), M16-1, -1, M16]], M16-1).
+
+ntoa([A | As], Max) ->
+    case
+        lists:all(
+          fun (X) when 0 =< X, X =< Max -> true;
+              (_) -> false
+          end, tuple_to_list(A))
+    of
+        true ->
+            S = inet:ntoa(A),
+            {ok, A} = inet:parse_address(S);
+        false ->
+            {error, einval} = inet:ntoa(A)
+    end,
+    ntoa(As, Max);
+ntoa([], _Max) ->
+    ok.
+
+rand_tuple(N, M) ->
+    rand_tuple(N, M, []).
+%%
+rand_tuple(0, _M, Acc) ->
+    list_to_tuple(Acc);
+rand_tuple(N, M, Acc) ->
+    rand_tuple(N - 1, M, [rand:uniform(M) - 1 | Acc]).
+
+
 t_gethostnative(Config) when is_list(Config) ->
     %% this will result in 26 bytes sent which causes problem in Windows
     %% if the port-program has not assured stdin to be read in BINARY mode
@@ -816,13 +977,11 @@ gethostnative_parallell(Config) when is_list(Config) ->
     end.
 
 do_gethostnative_parallell() ->
-    PA = filename:dirname(code:which(?MODULE)),
-    {ok,Node} = test_server:start_node(gethost_parallell, slave,
-				       [{args, "-pa " ++ PA}]),
+    {ok,Peer,Node} = ?CT_PEER(),
     ok = rpc:call(Node, ?MODULE, parallell_gethost, []),
     receive after 10000 -> ok end,
     pong = net_adm:ping(Node),
-    test_server:stop_node(Node),
+    peer:stop(Peer),
     ok.
 
 parallell_gethost() ->
@@ -953,9 +1112,7 @@ missing_hosts_reload(Config) when is_list(Config) ->
     ok = file:write_file(InetRc, "{hosts_file, \"" ++ HostsFile ++ "\"}.\n"),
     {error, enoent} = file:read_file_info(HostsFile),
     % start a node
-    Pa = filename:dirname(code:which(?MODULE)),
-    {ok, TestNode} = test_server:start_node(?MODULE, slave,
-        [{args, "-pa " ++ Pa ++ " -kernel inetrc '\"" ++ InetRc ++ "\"'"}]),
+    {ok, Peer, TestNode} = ?CT_PEER(["-kernel", "inetrc", "\"" ++ InetRc ++ "\""]),
     % ensure it has our RC
     Rc = rpc:call(TestNode, inet_db, get_rc, []),
     {hosts_file, HostsFile} = lists:keyfind(hosts_file, 1, Rc),
@@ -969,7 +1126,7 @@ missing_hosts_reload(Config) when is_list(Config) ->
     {ok,{hostent,"somehost",[],inet,4,[{1,2,3,4}]}} =
         rpc:call(TestNode, inet_hosts, gethostbyname, ["somehost"]),
     % cleanup
-    true = test_server:stop_node(TestNode).
+    peer:stop(Peer).
 
 
 %% The /etc/hosts file format and limitations is quite undocumented.
@@ -1047,9 +1204,7 @@ hosts_file_quirks(Config) when is_list(Config) ->
     ok = file:write_file(InetRc, "{hosts_file, \"" ++ HostsFile ++ "\"}.\n"),
     %%
     %% start a node
-    Pa = filename:dirname(code:which(?MODULE)),
-    {ok, TestNode} = test_server:start_node(?MODULE, slave,
-        [{args, "-pa " ++ Pa ++ " -kernel inetrc '\"" ++ InetRc ++ "\"'"}]),
+    {ok, Peer, TestNode} = ?CT_PEER(["-kernel", "inetrc", "\"" ++ InetRc ++ "\""]),
     %% ensure it has our RC
     Rc = rpc:call(TestNode, inet_db, get_rc, []),
     {hosts_file, HostsFile} = lists:keyfind(hosts_file, 1, Rc),
@@ -1096,7 +1251,7 @@ hosts_file_quirks(Config) when is_list(Config) ->
     hosts_file_quirks_verify(TestNode, V1),
     %%
     %% cleanup
-    true = test_server:stop_node(TestNode).
+    peer:stop(Peer).
 
 hosts_file_quirks_verify(_TestNode, Vs) ->
     hosts_file_quirks_verify(_TestNode, Vs, true).
@@ -1211,7 +1366,7 @@ gethostnative_control(Config, Opts) ->
 		    {skipped, "Not running native gethostbyname"}
 	    end;
 	_ ->
-	    {skipped, "Native not only lookup metod"}
+	    {skipped, "Native not only lookup method"}
     end.
 
 gethostnative_control_1(
@@ -1798,6 +1953,150 @@ add_del_host_v6(_Config) ->
     {ok, HostEnt} = inet_hosts:gethostbyname(Name, inet6).
 
 
+
+%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
+
+%%% This is a supremely simple test case.
+%%% It basically just tests that the function (inet:socknames/1,2)
+%%% returns with the expected result (a list of addresses, which in
+%%% the case of tcp and udp should be exactly one address long).
+
+socknames_sctp(Config) when is_list(Config) ->
+    %%% ?TC_TRY(?FUNCTION_NAME, fun() -> do_socknames_sctp(Config) end).
+    {skip, not_implemented_yet}.
+
+
+socknames_tcp(Config) when is_list(Config) ->
+    ?TC_TRY(?FUNCTION_NAME, fun() -> do_socknames_tcp0(Config) end).
+
+do_socknames_tcp0(_Config) ->
+    %% Begin with a the plain old boring (= port) socket(s)
+    ?P("Test socknames for 'old' socket (=port)"),
+    do_socknames_tcp1([]),
+
+    %% And *maybe* also check the 'new' shiny socket sockets
+    try socket:info() of
+        #{} ->
+            ?P("Test socknames for 'new' socket (=socket nif)"),
+            do_socknames_tcp1([{inet_backend, socket}])
+    catch
+        error:notsup ->
+            ?P("Skip test of socknames for 'new' socket (=socket nif)"),
+            ok;
+        error:undef:ST ->
+            case ST of
+                [{prim_socket,info,[],_}|_] ->
+                    ?P("Skip test of socknames for 'new' socket (=socket nif)")
+            end
+    end.
+
+
+do_socknames_tcp1(Conf) ->
+    ?P("try create listen socket"),
+    {ok, S1} = gen_tcp:listen(0, Conf),
+    ?P("try get socknames for (listen) socket: "
+       "~n      ~p", [S1]),
+    PortNumber1 = case inet:socknames(S1) of
+		      {ok, [{_Addr1, PortNo1}]} ->
+			  PortNo1;
+		      {ok, Addrs1} ->
+                          ?P("failed socknames: unexpected number of addresses"
+                             "~n      ~p", [Addrs1]),
+			  exit({unexpected_addrs_length, length(Addrs1)});
+		      {error, Reason1} ->
+                          ?P("failed socknames: error"
+                             "~n      ~p", [Reason1]),
+			  exit({skip, {listen_socket, Reason1}})
+		  end,
+    ?P("try connect to listen socket on port ~p", [PortNumber1]),
+    {ok, S2} = gen_tcp:connect("localhost", PortNumber1, Conf),
+    ?P("try get socknames for (connected) socket: "
+       "~n      ~p", [S2]),
+    case inet:socknames(S2) of
+	{ok, [_Addr2]} ->
+	    ok;
+	{ok, Addrs2} ->
+            ?P("failed socknames: unexpected number of addresses"
+               "~n      ~p", [Addrs2]),
+	    exit({unexpected_addrs_length, length(Addrs2)});
+	{error, Reason2} ->
+            ?P("failed socknames: error"
+               "~n      ~p", [Reason2]),
+	    exit({skip, {connected_socket, Reason2}})
+    end,
+    ?P("try accept connection"),
+    {ok, S3} = gen_tcp:accept(S1),
+    ?P("try get socknames for (accepted) socket: "
+       "~n      ~p", [S3]),
+    case inet:socknames(S3) of
+	{ok, [_Addr3]} ->
+	    ok;
+	{ok, Addrs3} ->
+            ?P("failed socknames: unexpected number of addresses"
+               "~n      ~p", [Addrs3]),
+	    exit({unexpected_addrs_length, length(Addrs3)});
+	{error, Reason3} ->
+            ?P("failed socknames: error"
+               "~n      ~p", [Reason3]),
+	    exit({skip, {accepted_socket, Reason3}})
+    end,
+    ?P("close socket(s)"),
+    (catch gen_tcp:close(S3)),
+    (catch gen_tcp:close(S2)),
+    (catch gen_tcp:close(S1)),
+    ?P("done"),
+    ok.
+
+
+socknames_udp(Config) when is_list(Config) ->
+    ?TC_TRY(?FUNCTION_NAME, fun() -> do_socknames_udp0(Config) end).
+
+do_socknames_udp0(_Config) ->
+    %% Begin with a the plain old boring (= port) socket(s)
+    ?P("Test socknames for 'old' socket (=port)"),
+    do_socknames_udp1([]),
+
+    %% And *maybe* also check the 'new' shiny socket sockets
+    try socket:info() of
+        #{} ->
+            ?P("Test socknames for 'new' socket (=socket nif)"),
+            do_socknames_udp1([{inet_backend, socket}])
+    catch
+        error : notsup ->
+            ?P("Skip test of socknames for 'new' socket (=socket nif)"),
+            ok;
+        error:undef:ST ->
+            case ST of
+                [{prim_socket,info,[],_}|_] ->
+                    ?P("Skip test of socknames for 'new' socket (=socket nif)")
+            end
+    end.
+
+
+do_socknames_udp1(Conf) ->
+    ?P("try create socket"),
+    {ok, S1} = gen_udp:open(0, Conf),
+    ?P("try get socknames for socket: "
+       "~n      ~p", [S1]),
+    case inet:socknames(S1) of
+        {ok, [_Addr1]} ->
+            ok;
+        {ok, Addrs1} ->
+            ?P("failed socknames: unexpected number of addresses"
+               "~n      ~p", [Addrs1]),
+            exit({unexpected_addrs_length, length(Addrs1)});
+        {error, Reason1} ->
+            ?P("failed socknames: error"
+               "~n      ~p", [Reason1]),
+            exit({skip, {listen_socket, Reason1}})
+    end,
+    ?P("close socket"),
+    (catch gen_udp:close(S1)),
+    ?P("done"),
+    ok.
+
+
+
 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
 pi(Item) ->
