@@ -1,7 +1,7 @@
 %%
 %% %CopyrightBegin%
 %%
-%% Copyright Ericsson AB 2019-2020. All Rights Reserved.
+%% Copyright Ericsson AB 2019-2023. All Rights Reserved.
 %%
 %% Licensed under the Apache License, Version 2.0 (the "License");
 %% you may not use this file except in compliance with the License.
@@ -22,6 +22,8 @@
 
 -module(ssl_cipher_suite_SUITE).
 
+-behaviour(ct_suite).
+
 -include_lib("common_test/include/ct.hrl").
 %% Callback functions
 -export([all/0,
@@ -122,10 +124,14 @@
          aes_128_gcm_sha256/1,
          chacha20_poly1305_sha256/1,
          aes_128_ccm_sha256/1,
-         aes_128_ccm_8_sha256/1
+         aes_128_ccm_8_sha256/1,
+         ecdhe_ecdsa_with_aes_128_ccm/1,
+         ecdhe_ecdsa_with_aes_256_ccm/1,
+         ecdhe_ecdsa_with_aes_128_ccm_8/1,
+         ecdhe_ecdsa_with_aes_256_ccm_8/1
         ]).
 
--define(TIMEOUT, {seconds, 5}).
+-define(TIMEOUT, {seconds, 10}).
 
 %%--------------------------------------------------------------------
 %% Common Test interface functions -----------------------------------
@@ -169,7 +175,11 @@ groups() ->
                        ecdhe_ecdsa_aes_128_gcm,
                        ecdhe_ecdsa_aes_256_cbc,
                        ecdhe_ecdsa_aes_256_gcm,
-                       ecdhe_ecdsa_chacha20_poly1305
+                       ecdhe_ecdsa_chacha20_poly1305,
+                       ecdhe_ecdsa_with_aes_128_ccm,
+                       ecdhe_ecdsa_with_aes_256_ccm,
+                       ecdhe_ecdsa_with_aes_128_ccm_8,
+                       ecdhe_ecdsa_with_aes_256_ccm_8
                       ]},
      {rsa, [], [rsa_3des_ede_cbc, 
                 rsa_aes_128_cbc,
@@ -376,7 +386,7 @@ init_per_testcase(TestCase, Config) when TestCase == psk_3des_ede_cbc;
                                          TestCase == ecdh_anon_3des_ede_cbc;
                                          TestCase == ecdhe_ecdsa_3des_ede_cbc ->
     SupCiphers = proplists:get_value(ciphers, crypto:supports()),
-    case lists:member(des_ede3, SupCiphers) of
+    case lists:member(des_ede3_cbc, SupCiphers) of
         true ->
             ct:timetrap(?TIMEOUT),
             Config;
@@ -480,6 +490,26 @@ init_per_testcase(aes_128_ccm_8_sha256, Config) ->
         _ ->
             {skip, "Missing AES_128_CCM_8_SHA256 crypto support"}
     end;
+init_per_testcase(TestCase, Config) when TestCase == ecdhe_ecdsa_with_aes_128_ccm;
+                                         TestCase == ecdhe_ecdsa_with_aes_128_ccm_8->
+    SupCiphers = proplists:get_value(ciphers, crypto:supports()),
+    case lists:member(aes_128_ccm, SupCiphers) of
+        true ->
+            ct:timetrap(?TIMEOUT),
+            Config;
+        _ ->
+            {skip, "Missing AES_128_CCM crypto support"}
+    end;
+init_per_testcase(TestCase, Config) when TestCase == ecdhe_ecdsa_with_aes_256_ccm;
+                                         TestCase == ecdhe_ecdsa_with_aes_256_ccm_8 ->
+    SupCiphers = proplists:get_value(ciphers, crypto:supports()),
+    case lists:member(aes_256_ccm, SupCiphers) of
+        true ->
+            ct:timetrap(?TIMEOUT),
+            Config;
+        _ ->
+            {skip, "Missing AES_256_CCM crypto support"}
+    end;
 init_per_testcase(TestCase, Config) ->
     Cipher = ssl_test_lib:test_cipher(TestCase, Config),
     SupCiphers = proplists:get_value(ciphers, crypto:supports()),
@@ -500,57 +530,75 @@ end_per_testcase(_TestCase, Config) ->
 init_certs(srp_rsa, Config) ->
     DefConf = ssl_test_lib:default_cert_chain_conf(),
     CertChainConf = ssl_test_lib:gen_conf(rsa, rsa, DefConf, DefConf),
-    #{server_config := ServerOpts,
-      client_config := ClientOpts} 
+    #{server_config := ServerOpts0,
+      client_config := ClientOpts0}
         = public_key:pkix_test_data(CertChainConf),
-    [{tls_config, #{server_config => [{user_lookup_fun, {fun ssl_test_lib:user_lookup/3, undefined}} | ServerOpts],
-                    client_config => [{srp_identity, {"Test-User", "secret"}} | ClientOpts]}} |
+    Version = ssl_test_lib:n_version(proplists:get_value(version, Config)),
+    ServerOpts = ssl_test_lib:sig_algs(rsa, Version) ++  ServerOpts0,
+    ClientOpts = ssl_test_lib:sig_algs(rsa, Version) ++ ClientOpts0,
+    [{tls_config, #{server_config =>
+                        [{user_lookup_fun, {fun ssl_test_lib:user_lookup/3, undefined}} | ServerOpts],
+                    client_config =>
+                        [{srp_identity, {"Test-User", "secret"}}, {verify, verify_none} | ClientOpts]}} |
      proplists:delete(tls_config, Config)];
 init_certs(srp_anon, Config) ->
     [{tls_config, #{server_config => [{user_lookup_fun, {fun ssl_test_lib:user_lookup/3, undefined}}],
-                    client_config => [{srp_identity, {"Test-User", "secret"}}]}} |
+                    client_config => [{srp_identity, {"Test-User", "secret"}}, {verify, verify_none}]}} |
      proplists:delete(tls_config, Config)];
 init_certs(rsa_psk, Config) ->
     ClientExt = x509_test:extensions([{key_usage, [digitalSignature, keyEncipherment]}]),
-    {ClientOpts, ServerOpts} = ssl_test_lib:make_rsa_cert_chains([{server_chain, 
-                                                                   [[],[],[{extensions, ClientExt}]]}], 
+    {ClientOpts, ServerOpts} = ssl_test_lib:make_rsa_cert_chains([{server_chain,
+                                                                   [[],[],[{extensions, ClientExt}]]}],
                                                                  Config, "_peer_keyEncipherment"),
     PskSharedSecret = <<1,2,3,4,5,6,7,8,9,10,11,12,13,14,15>>,
     [{tls_config, #{server_config => [{user_lookup_fun, {fun ssl_test_lib:user_lookup/3, PskSharedSecret}} | ServerOpts],
                     client_config => [{psk_identity, "Test-User"},
-                                      {user_lookup_fun, {fun ssl_test_lib:user_lookup/3, PskSharedSecret}} | ClientOpts]}} |
+                                      {user_lookup_fun, {fun ssl_test_lib:user_lookup/3, PskSharedSecret}},
+                                      {verify, verify_none} | ClientOpts]}} |
      proplists:delete(tls_config, Config)];
 init_certs(rsa, Config) ->
     ClientExt = x509_test:extensions([{key_usage, [digitalSignature, keyEncipherment]}]),
-    {ClientOpts, ServerOpts} = ssl_test_lib:make_rsa_cert_chains([{server_chain, 
-                                                                   [[],[],[{extensions, ClientExt}]]}], 
+    {ClientOpts0, ServerOpts0} = ssl_test_lib:make_rsa_cert_chains([{server_chain,
+                                                                   [[],[],[{extensions, ClientExt}]]}],
                                                                  Config, "_peer_keyEncipherment"),
+    Version = ssl_test_lib:n_version(proplists:get_value(version, Config)),
+    ServerOpts = ssl_test_lib:sig_algs(rsa, Version) ++  ServerOpts0,
+    ClientOpts = ssl_test_lib:sig_algs(rsa, Version) ++ ClientOpts0,
     [{tls_config, #{server_config => ServerOpts,
                     client_config => ClientOpts}} |
      proplists:delete(tls_config, Config)];
 init_certs(ecdhe_1_3_rsa_cert, Config) ->
     ClientExt = x509_test:extensions([{key_usage, [digitalSignature]}]),
-    {ClientOpts, ServerOpts} = ssl_test_lib:make_rsa_cert_chains([{server_chain, 
-                                                                   [[],[],[{extensions, ClientExt}]]}], 
+    {ClientOpts0, ServerOpts0} = ssl_test_lib:make_rsa_cert_chains([{server_chain,
+                                                                   [[],[],[{extensions, ClientExt}]]}],
                                                                  Config, "_peer_rsa_digitalsign"),
+    Version = ssl_test_lib:n_version(proplists:get_value(version, Config)),
+    ServerOpts = ssl_test_lib:sig_algs(rsa, Version) ++  ServerOpts0,
+    ClientOpts = ssl_test_lib:sig_algs(rsa, Version) ++ ClientOpts0,
     [{tls_config, #{server_config => ServerOpts,
                     client_config => ClientOpts}} |
      proplists:delete(tls_config, Config)];
 init_certs(dhe_dss, Config) ->
     DefConf = ssl_test_lib:default_cert_chain_conf(),
-    CertChainConf = ssl_test_lib:gen_conf(dsa, dsa, DefConf, DefConf),
-    #{server_config := ServerOpts,
-      client_config := ClientOpts} 
+    CertChainConf = ssl_test_lib:gen_conf(dsa, dsa, DefConf, DefConf), 
+    #{server_config := ServerOpts0,
+      client_config := ClientOpts0}
         = public_key:pkix_test_data(CertChainConf),
+    Version = ssl_test_lib:n_version(proplists:get_value(version, Config)),
+    ServerOpts = ssl_test_lib:sig_algs(dsa, Version) ++  ServerOpts0,
+    ClientOpts = ssl_test_lib:sig_algs(dsa, Version) ++ ClientOpts0,
     [{tls_config, #{server_config => ServerOpts,
                     client_config => ClientOpts}} |
      proplists:delete(tls_config, Config)];
 init_certs(srp_dss, Config) ->
     DefConf = ssl_test_lib:default_cert_chain_conf(),
     CertChainConf = ssl_test_lib:gen_conf(dsa, dsa, DefConf, DefConf),
-    #{server_config := ServerOpts,
-      client_config := ClientOpts} 
+    #{server_config := ServerOpts0,
+      client_config := ClientOpts0}
         = public_key:pkix_test_data(CertChainConf),
+    Version = ssl_test_lib:n_version(proplists:get_value(version, Config)),
+    ServerOpts = ssl_test_lib:sig_algs(dsa, Version) ++  ServerOpts0,
+    ClientOpts = ssl_test_lib:sig_algs(dsa, Version) ++ ClientOpts0,
     [{tls_config, #{server_config => [{user_lookup_fun, {fun ssl_test_lib:user_lookup/3, undefined}} | ServerOpts],
                     client_config => [{srp_identity, {"Test-User", "secret"}} | ClientOpts]}} |
        proplists:delete(tls_config, Config)];
@@ -558,9 +606,12 @@ init_certs(GroupName, Config) when GroupName == dhe_rsa;
                                    GroupName == ecdhe_rsa ->
     DefConf = ssl_test_lib:default_cert_chain_conf(),
     CertChainConf = ssl_test_lib:gen_conf(rsa, rsa, DefConf, DefConf),
-    #{server_config := ServerOpts,
-      client_config := ClientOpts} 
+    #{server_config := ServerOpts0,
+      client_config := ClientOpts0}
         = public_key:pkix_test_data(CertChainConf),
+    Version = ssl_test_lib:n_version(proplists:get_value(version, Config)),
+    ServerOpts = ssl_test_lib:sig_algs(rsa, Version) ++  ServerOpts0,
+    ClientOpts = ssl_test_lib:sig_algs(rsa, Version) ++ ClientOpts0,
     [{tls_config, #{server_config => ServerOpts,
                     client_config => ClientOpts}} |
      proplists:delete(tls_config, Config)];
@@ -568,9 +619,12 @@ init_certs(GroupName, Config) when GroupName == dhe_ecdsa;
                                    GroupName == ecdhe_ecdsa ->
     DefConf = ssl_test_lib:default_cert_chain_conf(),
     CertChainConf = ssl_test_lib:gen_conf(ecdsa, ecdsa, DefConf, DefConf),
-    #{server_config := ServerOpts,
-      client_config := ClientOpts} 
+    #{server_config := ServerOpts0,
+      client_config := ClientOpts0}
         = public_key:pkix_test_data(CertChainConf),
+    Version = ssl_test_lib:n_version(proplists:get_value(version, Config)),
+    ServerOpts = ssl_test_lib:sig_algs(ecdsa, Version) ++  ServerOpts0,
+    ClientOpts = ssl_test_lib:sig_algs(ecdsa, Version) ++ ClientOpts0,
     [{tls_config, #{server_config => ServerOpts,
                     client_config => ClientOpts}} |
      proplists:delete(tls_config, Config)];
@@ -579,17 +633,17 @@ init_certs(GroupName, Config) when GroupName == psk;
                                    GroupName == ecdhe_psk ->
     PskSharedSecret = <<1,2,3,4,5,6,7,8,9,10,11,12,13,14,15>>,
     [{tls_config, #{server_config => [{user_lookup_fun, {fun ssl_test_lib:user_lookup/3, PskSharedSecret}}],
-                    client_config => [{psk_identity, "Test-User"},
+                    client_config => [{verify, verify_none}, {psk_identity, "Test-User"},
                                       {user_lookup_fun, {fun ssl_test_lib:user_lookup/3, PskSharedSecret}}]}} |
-     proplists:delete(tls_config, Config)]; 
-init_certs(srp, Config) -> 
+     proplists:delete(tls_config, Config)];
+init_certs(srp, Config) ->
       [{tls_config, #{server_config => [{user_lookup_fun, {fun ssl_test_lib:user_lookup/3, undefined}}],
-                      client_config => [{srp_identity, {"Test-User", "secret"}}]}} |
+                      client_config => [{verify, verify_none},{srp_identity, {"Test-User", "secret"}}]}} |
        proplists:delete(tls_config, Config)];
-init_certs(_GroupName, Config) -> 
+init_certs(_GroupName, Config) ->
     %% Anonymous does not need certs
-     [{tls_config, #{server_config => [],
-                     client_config => []}} |
+     [{tls_config, #{server_config => [{verify, verify_none}],
+                     client_config => [{verify, verify_none}]}} |
        proplists:delete(tls_config, Config)].
 
 %%--------------------------------------------------------------------
@@ -742,6 +796,18 @@ ecdhe_ecdsa_aes_256_gcm(Config) when is_list(Config) ->
 
 ecdhe_ecdsa_chacha20_poly1305(Config) when is_list(Config) ->
     run_ciphers_test(ecdhe_ecdsa, 'chacha20_poly1305', Config).
+
+ecdhe_ecdsa_with_aes_128_ccm(Config) when is_list(Config) ->
+    run_ciphers_test(ecdhe_ecdsa, 'aes_128_ccm', Config). 
+
+ecdhe_ecdsa_with_aes_256_ccm(Config) when is_list(Config) ->
+    run_ciphers_test(ecdhe_ecdsa, 'aes_256_ccm', Config). 
+
+ecdhe_ecdsa_with_aes_128_ccm_8(Config) when is_list(Config) ->
+    run_ciphers_test(ecdhe_ecdsa, 'aes_128_ccm_8', Config). 
+
+ecdhe_ecdsa_with_aes_256_ccm_8(Config) when is_list(Config) ->
+    run_ciphers_test(ecdhe_ecdsa, 'aes_256_ccm_8', Config). 
 %%--------------------------------------------------------------------
 %% DHE_DSS --------------------------------------------------------
 %%--------------------------------------------------------------------
@@ -936,14 +1002,14 @@ cipher_suite_test(ErlangCipherSuite, Version, Config) ->
 
 
 test_ciphers(Kex, Cipher, Version) ->
-    ssl:filter_cipher_suites(ssl:cipher_suites(all, Version) ++ ssl:cipher_suites(anonymous, Version), 
-                             [{key_exchange, 
-                               fun(Kex0) when Kex0 == Kex -> true; 
-                                  (_) -> false 
-                               end}, 
-                              {cipher,  
-                               fun(Cipher0) when Cipher0 == Cipher -> true; 
-                                  (_) -> false 
+    ssl:filter_cipher_suites(ssl:cipher_suites(all, Version) ++ ssl:cipher_suites(anonymous, Version),
+                             [{key_exchange,
+                               fun(Kex0) when Kex0 == Kex -> true;
+                                  (_) -> false
+                               end},
+                              {cipher,
+                               fun(Cipher0) when Cipher0 == Cipher -> true;
+                                  (_) -> false
                                end}]).
 
 
