@@ -1,7 +1,7 @@
 %%
 %% %CopyrightBegin%
 %%
-%% Copyright Ericsson AB 2008-2020. All Rights Reserved.
+%% Copyright Ericsson AB 2008-2023. All Rights Reserved.
 %%
 %% Licensed under the Apache License, Version 2.0 (the "License");
 %% you may not use this file except in compliance with the License.
@@ -21,6 +21,8 @@
 %%
 -module(ssl_npn_SUITE).
 
+-behaviour(ct_suite).
+
 -include_lib("common_test/include/ct.hrl").
 
 %% Callback functions
@@ -66,14 +68,18 @@
 all() ->
     [{group, 'tlsv1.2'},
      {group, 'tlsv1.1'},
-     {group, 'tlsv1'}
+     {group, 'tlsv1'},
+     {group, 'dtlsv1.2'},
+     {group, 'dtlsv1'}
     ].
 
 groups() ->
     [
      {'tlsv1.2', [], next_protocol_tests()},
      {'tlsv1.1', [], next_protocol_tests()},
-     {'tlsv1', [], next_protocol_tests()}
+     {'tlsv1', [], next_protocol_tests()},
+     {'dtlsv1.2', [], next_protocol_tests()},
+     {'dtlsv1', [], next_protocol_tests()}
     ].
 
 next_protocol_tests() ->
@@ -115,7 +121,8 @@ end_per_group(GroupName, Config) ->
 
 init_per_testcase(_TestCase, Config) ->
     ssl_test_lib:ct_log_supported_protocol_versions(Config),
-    ct:log("Ciphers: ~p~n ", [ ssl:cipher_suites()]),
+    Version = proplists:get_value(version, Config),
+    ct:log("Ciphers: ~p~n ", [ ssl:cipher_suites(default, Version)]),
     ct:timetrap(?TIMEOUT),
     Config.
 
@@ -132,10 +139,11 @@ validate_empty_protocols_are_not_allowed(Config) when is_list(Config) ->
 			    [{next_protocols_advertised, [<<"foo/1">>, <<"">>]}])),
     {error, {options, {client_preferred_next_protocols, {invalid_protocol, <<>>}}}}
 	= (catch ssl:connect({127,0,0,1}, 9443,
-			     [{client_preferred_next_protocols,
+			     [{verify, verify_none}, {client_preferred_next_protocols,
 			       {client, [<<"foo/1">>, <<"">>], <<"foox/1">>}}], infinity)),
     Option = {client_preferred_next_protocols, {invalid_protocol, <<"">>}},
-    {error, {options, Option}} = (catch ssl:connect({127,0,0,1}, 9443, [Option], infinity)).
+    {error, {options, Option}} = (catch ssl:connect({127,0,0,1}, 9443, 
+                                                    [{verify, verify_none}, Option], infinity)).
 
 %--------------------------------------------------------------------------------
 
@@ -147,12 +155,13 @@ validate_empty_advertisement_list_is_allowed(Config) when is_list(Config) ->
 
 validate_advertisement_must_be_a_binary_list(Config) when is_list(Config) ->
     Option = {next_protocols_advertised, blah},
-    {error, {options, Option}} = (catch ssl:listen(9443, [Option])).
+    {error, {options, Option}} = (catch ssl:listen(9443, [{verify, verify_none}, Option])).
 %--------------------------------------------------------------------------------
 
 validate_client_protocols_must_be_a_tuple(Config) when is_list(Config)  ->
     Option = {client_preferred_next_protocols, [<<"foo/1">>]},
-    {error, {options, Option}} = (catch ssl:connect({127,0,0,1}, 9443, [Option])).
+    {error, {options, Option}} = (catch ssl:connect({127,0,0,1}, 9443,
+                                                    [{verify, verify_none}, Option])).
 
 %--------------------------------------------------------------------------------
 
