@@ -1,7 +1,7 @@
 %%
 %% %CopyrightBegin%
 %%
-%% Copyright Ericsson AB 2008-2020. All Rights Reserved.
+%% Copyright Ericsson AB 2008-2023. All Rights Reserved.
 %%
 %% Licensed under the Apache License, Version 2.0 (the "License");
 %% you may not use this file except in compliance with the License.
@@ -20,9 +20,6 @@
 
 -module(tls_1_3_record_SUITE).
 
-%% Note: This directive should only be used in test suites.
--compile(export_all).
-
 -include_lib("common_test/include/ct.hrl").
 -include_lib("ssl/src/tls_record.hrl").
 -include_lib("ssl/src/tls_handshake.hrl").
@@ -30,6 +27,26 @@
 -include_lib("ssl/src/ssl_cipher.hrl").
 -include_lib("ssl/src/ssl_internal.hrl").
 
+%% Callback functions
+-export([all/0,
+         init_per_suite/1,
+         end_per_suite/1]).
+
+%% Testcases
+-export([encode_decode/0,
+         encode_decode/1,
+         finished_verify_data/0,
+         finished_verify_data/1,
+         '1_RTT_handshake'/0,
+         '1_RTT_handshake'/1,
+         '0_RTT_handshake'/0,
+         '0_RTT_handshake'/1
+        ]).
+
+%%--------------------------------------------------------------------
+%% Common Test interface functions -----------------------------------
+%%--------------------------------------------------------------------
+
 all() ->
     [encode_decode,
      finished_verify_data,
@@ -73,20 +90,25 @@ encode_decode(_Config) ->
                 client_verify_data => undefined,compression_state => undefined,
                 mac_secret => undefined,secure_renegotiation => undefined,
                 security_parameters =>
-                    {security_parameters,
-                     <<19,2>>,
-                     0,8,2,undefined,undefined,undefined,undefined,undefined,
-                     sha384,undefined,undefined,
-                     {handshake_secret,
-                      <<128,229,186,211,62,127,182,20,62,166,233,23,135,64,121,
-                        3,104,251,214,161,253,31,3,2,232,37,8,221,189,72,64,218,
-                        121,41,112,148,254,34,68,164,228,60,161,201,132,55,56,
-                        157>>}, undefined, undefined,
-                     undefined,
-                     <<92,24,205,75,244,60,136,212,250,32,214,20,37,3,213,87,61,207,
-                       147,61,168,145,177,118,160,153,33,53,48,108,191,174>>,
-                     undefined},
-                sequence_number => 0,server_verify_data => undefined},
+                    #security_parameters{
+                       cipher_suite          = <<19,2>>,
+                       connection_end        = 0,
+                       bulk_cipher_algorithm = 8,
+                       cipher_type           = 2,
+                       prf_algorithm         = sha384,
+                       master_secret         =
+                           {handshake_secret,
+                            <<128,229,186,211,62,127,182,20,62,166,233,23,135,64,121,
+                              3,104,251,214,161,253,31,3,2,232,37,8,221,189,72,64,218,
+                              121,41,112,148,254,34,68,164,228,60,161,201,132,55,56,
+                              157>>},
+                       server_random         =
+                           <<92,24,205,75,244,60,136,212,250,32,214,20,37,3,213,87,61,207,
+                             147,61,168,145,177,118,160,153,33,53,48,108,191,174>>},
+                sequence_number => 0,server_verify_data => undefined,
+                pending_early_data_size => 0,
+                trial_decryption => false,
+                early_data_accepted => false},
           current_write =>
               #{beast_mitigation => one_n_minus_one,
                 cipher_state =>
@@ -99,19 +121,21 @@ encode_decode(_Config) ->
                 client_verify_data => undefined,compression_state => undefined,
                 mac_secret => undefined,secure_renegotiation => undefined,
                 security_parameters =>
-                    {security_parameters,
-                     <<19,2>>,
-                     0,8,2,undefined,undefined,undefined,undefined,undefined,
-                     sha384,undefined,undefined,
-                     {handshake_secret,
-                      <<128,229,186,211,62,127,182,20,62,166,233,23,135,64,121,
-                        3,104,251,214,161,253,31,3,2,232,37,8,221,189,72,64,218,
-                        121,41,112,148,254,34,68,164,228,60,161,201,132,55,56,
-                        157>>}, undefined, undefined,
-                     undefined,
-                     <<92,24,205,75,244,60,136,212,250,32,214,20,37,3,213,87,61,207,
-                       147,61,168,145,177,118,160,153,33,53,48,108,191,174>>,
-                     undefined},
+                    #security_parameters{
+                       cipher_suite          = <<19,2>>,
+                       connection_end        = 0,
+                       bulk_cipher_algorithm = 8,
+                       cipher_type           = 2,
+                       prf_algorithm         = sha384,
+                       master_secret         =
+                           {handshake_secret,
+                            <<128,229,186,211,62,127,182,20,62,166,233,23,135,64,121,
+                              3,104,251,214,161,253,31,3,2,232,37,8,221,189,72,64,218,
+                              121,41,112,148,254,34,68,164,228,60,161,201,132,55,56,
+                              157>>},
+                     server_random         =
+                           <<92,24,205,75,244,60,136,212,250,32,214,20,37,3,213,87,61,207,
+                             147,61,168,145,177,118,160,153,33,53,48,108,191,174>>},
                 sequence_number => 0,server_verify_data => undefined},max_fragment_length => undefined},
 
     PlainText = [11,
@@ -150,9 +174,9 @@ encode_decode(_Config) ->
                    146,152,146,151,107,126,216,210,9,93,0,0>>],
 
     {[_Header|Encoded], _} = tls_record_1_3:encode_plain_text(22, PlainText, ConnectionStates),
-    CipherText = #ssl_tls{type = 23, version = {3,3}, fragment = Encoded},
+    CipherText = #ssl_tls{type = 23, version = ?TLS_1_2, fragment = Encoded},
 
-    {#ssl_tls{type = 22, version = {3,4}, fragment = DecodedText}, _} =
+    {#ssl_tls{type = 22, version = ?TLS_1_3, fragment = DecodedText}, _} =
         tls_record_1_3:decode_cipher_text(CipherText, ConnectionStates),
 
     DecodedText = iolist_to_binary(PlainText),
@@ -236,7 +260,7 @@ encode_decode(_Config) ->
           01 04 02 05 02 06 02 02 02 00 2d 00 02 01 01 00 1c 00 02 40 01"),
 
     {CHEncrypted, _} =
-	tls_record:encode_handshake(ClientHello, {3,4}, ConnStatesNull),
+	tls_record:encode_handshake(ClientHello, ?TLS_1_3, ConnStatesNull),
     ClientHelloRecord = iolist_to_binary(CHEncrypted),
 
     %% {server}  extract secret "early":
@@ -250,7 +274,7 @@ encode_decode(_Config) ->
     %%       e2 10 ad f3 00 aa 1f 26 60 e1 b2 2e 10 f1 70 f9 2a
     HKDFAlgo = sha256,
     Salt = binary:copy(<<?BYTE(0)>>, 32),
-    IKM = binary:copy(<<?BYTE(0)>>, 32),
+    _IKM = binary:copy(<<?BYTE(0)>>, 32),
     EarlySecret =
         hexstr2bin("33 ad 0a 1c 60 7e c0 3b 09 e6 cd 98 93 68 0c
           e2 10 ad f3 00 aa 1f 26 60 e1 b2 2e 10 f1 70 f9 2a"),
@@ -279,7 +303,7 @@ encode_decode(_Config) ->
         hexstr2bin("b1 58 0e ea df 6d d5 89 b8 ef 4f 2d 56
          52 57 8c c8 10 e9 98 01 91 ec 8d 05 83 08 ce a2 16 a2 1e"),
 
-    SPublicKey =
+    _SPublicKey =
         hexstr2bin("c9 82 88 76 11 20 95 fe 66 76 2b db f7 c6
          72 e1 56 d6 cc 25 3b 83 3d f1 dd 69 b1 b0 4e 75 1f 0f"),
 
@@ -491,7 +515,7 @@ encode_decode(_Config) ->
           cc 25 3b 83 3d f1 dd 69 b1 b0 4e 75 1f 0f 00 2b 00 02 03 04"),
 
     {SHEncrypted, _} =
-	tls_record:encode_handshake(ServerHello, {3,4}, ConnStatesNull),
+	tls_record:encode_handshake(ServerHello, ?TLS_1_3, ConnStatesNull),
     ServerHelloRecord = iolist_to_binary(SHEncrypted),
 
     %% {server}  derive write traffic keys for handshake data:
@@ -527,7 +551,8 @@ encode_decode(_Config) ->
     %% TODO: remove hardcoded IV size
     WriteIVInfo = tls_v1:create_info(<<"iv">>, <<>>,  12),
 
-    {WriteKey, WriteIV} = tls_v1:calculate_traffic_keys(HKDFAlgo, Cipher, SHSTrafficSecret),
+    KeyLength = ssl_cipher:key_material(Cipher),
+    {WriteKey, WriteIV} = tls_v1:calculate_traffic_keys(HKDFAlgo, KeyLength, SHSTrafficSecret),
 
     %% {server}  construct an EncryptedExtensions handshake message:
     %%
@@ -660,7 +685,7 @@ encode_decode(_Config) ->
 
     FinishedHS = #finished{verify_data = FinishedVerifyData},
 
-    FinishedIOList = tls_handshake:encode_handshake(FinishedHS, {3,4}),
+    FinishedIOList = tls_handshake:encode_handshake(FinishedHS, ?TLS_1_3),
     FinishedHSBin = iolist_to_binary(FinishedIOList),
 
     %% {server}  derive secret "tls13 c ap traffic":
@@ -807,7 +832,7 @@ encode_decode(_Config) ->
     SWIV =
         hexstr2bin("cf 78 2b 88 dd 83 54 9a ad f1 e9 84"),
 
-    {SWKey, SWIV} = tls_v1:calculate_traffic_keys(HKDFAlgo, Cipher, SAPTrafficSecret),
+    {SWKey, SWIV} = tls_v1:calculate_traffic_keys(HKDFAlgo, KeyLength, SAPTrafficSecret),
 
     %% {server}  derive read traffic keys for handshake data:
     %%
@@ -832,7 +857,7 @@ encode_decode(_Config) ->
     SRIV =
         hexstr2bin("5b d3 c7 1b 83 6e 0b 76 bb 73 26 5f"),
 
-    {SRKey, SRIV} = tls_v1:calculate_traffic_keys(HKDFAlgo, Cipher, CHSTrafficSecret),
+    {SRKey, SRIV} = tls_v1:calculate_traffic_keys(HKDFAlgo, KeyLength, CHSTrafficSecret),
 
     %% {client}  calculate finished "tls13 finished":
     %%
@@ -882,7 +907,7 @@ encode_decode(_Config) ->
 
     CFinished = #finished{verify_data = CFinishedVerifyData},
 
-    CFinishedIOList = tls_handshake:encode_handshake(CFinished, {3,4}),
+    CFinishedIOList = tls_handshake:encode_handshake(CFinished, ?TLS_1_3),
     CFinishedBin = iolist_to_binary(CFinishedIOList),
 
     %% {client}  derive write traffic keys for application data:
@@ -909,7 +934,7 @@ encode_decode(_Config) ->
     CWIV =
         hexstr2bin("5b 78 92 3d ee 08 57 90 33 e5 23 d9"),
 
-    {CWKey, CWIV} = tls_v1:calculate_traffic_keys(HKDFAlgo, Cipher, CAPTrafficSecret),
+    {CWKey, CWIV} = tls_v1:calculate_traffic_keys(HKDFAlgo, KeyLength, CAPTrafficSecret),
 
     %% {client}  derive secret "tls13 res master":
     %%
@@ -1029,7 +1054,7 @@ encode_decode(_Config) ->
        ticket_nonce = Nonce,
        ticket = Ticket,
        extensions = _Extensions
-      } = tls_handshake:decode_handshake({3,4}, NWT, TicketBody),
+      } = tls_handshake:decode_handshake(?TLS_1_3, NWT, TicketBody),
 
     %% ResPRK = resumption master secret
     ResExpanded = tls_v1:pre_shared_key(ResPRK, Nonce, HKDFAlgo),
@@ -1263,7 +1288,7 @@ encode_decode(_Config) ->
 
     <<?BYTE(CH), ?UINT24(_Length), ClientHelloBody/binary>> = ClientHelloRecord,
     #client_hello{extensions = #{pre_shared_key := PreSharedKey}} =
-        tls_handshake:decode_handshake({3,4}, CH, ClientHelloBody),
+        tls_handshake:decode_handshake(?TLS_1_3, CH, ClientHelloBody),
 
     #pre_shared_key_client_hello{
        offered_psks = #offered_psks{
@@ -1405,7 +1430,7 @@ finished_verify_data(_Config) ->
 
 hexstr2int(S) ->
     B = hexstr2bin(S),
-    Bits = size(B) * 8,
+    Bits = byte_size(B) * 8,
     <<Integer:Bits/integer>> = B,
     Integer.
 
